# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Open a new viewport in Rhino that shows the parallel-projected view from the sun.
_
This is useful for understanding what parts of Rhino geometry are shaded at a
particular hour of the day.
-

    Args:
        _vector: A sun vector from which the the Rhino view will be generated.
            Use the "LB SunPath" component to generate sun vectors.
        _center_pt_: The target point of the camera for the Rhino view that will be
            generated.  This point should be close to the Rhino geometry that
            you are interested in viewing from the sun. If no point is provided,
            the Rhino origin will be used (0, 0, 0).
        width_: An optional interger for the width (in pixels) of the Rhino
            viewport that will be generated.
        height_: An optional interger for the height (in pixels) of the Rhino
            viewport that will be generated.
        mode_: An optional text input for the display mode of the Rhino viewport
            that will be generated. For example: Wireframe, Shaded, Rendered, etc.

    Returns:
        report: ...
"""

ghenv.Component.Name = 'LB View From Sun'
ghenv.Component.NickName = 'ViewFromSun'
ghenv.Component.Message = '1.6.1'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '3 :: Analyze Geometry'
ghenv.Component.AdditionalHelpFromDocStrings = '3'

try:
    from ladybug_rhino.viewport import viewport_by_name, open_viewport, \
        set_iso_view_direction, set_view_display_mode
    from ladybug_rhino.grasshopper import all_required_inputs, component_guid
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    view_name = 'ViewFromSun_{}'.format(component_guid(ghenv.Component))
    print(view_name)  # print so that the user has the name if needed

    # get the viewport from which the direction will be set
    view_port = None
    if not width_ and not height_:  # no need to generate new view; get existing one
        try:
            view_port = viewport_by_name(view_name)
        except ValueError:  # the viewport does not yet exist
            pass
    if view_port is None:
        view_port = open_viewport(view_name, width_, height_)

    # set the direction of the viewport camera
    set_iso_view_direction(view_port, _vector, _center_pt_)

    # set the display mode if requested
    if mode_:
        set_view_display_mode(view_port, mode_)



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Visualize a sky matrix from the "LB Cumulative Sky Matrix" component as a colored
dome, subdivided into patches with a radiation value for each patch.
-

    Args:
        _sky_mtx: A Sky Matrix from the "LB Cumulative Sky Matrix" component or the
            "LB Benefit Sky Matrix" component, which describes the radiation
            coming from the various patches of the sky.
        _center_pt_: A point for the center of the dome. (Default: (0, 0, 0))
        _scale_: A number to set the scale of the sky dome. The default is 1,
            which corresponds to a radius of 100 meters in the current Rhino
            model's unit system.
        projection_: Optional text for the name of a projection to use from the sky
            dome hemisphere to the 2D plane. If None, a 3D sky dome will be drawn
            instead of a 2D one. (Default: None) Choose from the following:
                * Orthographic
                * Stereographic
        irradiance_: Boolean to note whether the sky dome should be plotted with units of
            cumulative Radiation (kWh/m2) [False] or with units of average
            Irradiance (W/m2) [True]. (Default: False).
        show_comp_: Boolean to indicate whether only one dome with total radiation
            should be displayed (False) or three domes with the solar radiation
            components (total, direct, and diffuse) should be shown. (Default: False).
        legend_par_: An optional LegendParameter object to change the display of the
            sky dome (Default: None).

    Returns:
        report: ...
        mesh:  A colored mesh representing the intensity of radiation for each of
            the sky patches within the sky dome.
        compass: A set of circles, lines and text objects that mark the cardinal
            directions in relation to the sky dome.
        legend: A legend showing the kWh/m2 or W/m2 values that correspond to the colors
            of the mesh.
        title: A text object for the title of the sky dome.
        patch_vecs: A list of vectors for each of the patches of the sky dome.
            All vectors are unit vectors and point from the center towards each
            of the patches. They can be used to construct visualizations of the
            rays used to perform radiation analysis.
        patch_values: Radiation values for each of the sky patches in kWh/m2 or W/m2.
            This will be one list if show_comp_ is "False" and a list of 3
            lists (aka. a Data Tree) for total, direct, diffuse if show_comp_
            is "True".
        mesh_values: Radiation values for each face of the dome mesh in kWh/m2. This
            can be used to post-process the radiation data and then regenerate
            the dome visualization using the mesh output from this component
            and the "LB Spatial Heatmap" component. Examples of useful post-
            processing include converting the units to something other than
            kWh/m2, inverting the +/- sign of radiation values depending on
            whether radiation is helpful or harmful to building thermal loads,
            etc. This will be one list if show_comp_ is "False" and a list of
            3 lists (aka. a Data Tree) for total, direct, diffuse if show_comp_
            is "True".
        vis_set: An object containing VisualizationSet arguments for drawing a detailed
            version of the Sky Dome in the Rhino scene. This can be connected to
            the "LB Preview Visualization Set" component to display this version
            of the Sky Dome in Rhino.
"""

ghenv.Component.Name = 'LB Sky Dome'
ghenv.Component.NickName = 'SkyDome'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '2 :: Visualize Data'
ghenv.Component.AdditionalHelpFromDocStrings = '3'

try:
    from ladybug_geometry.geometry3d.pointvector import Point3D
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_geometry:\n\t{}'.format(e))

try:
    from ladybug_radiance.visualize.skydome import SkyDome
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_radiance:\n\t{}'.format(e))

try:
    from ladybug_rhino.config import conversion_to_meters
    from ladybug_rhino.togeometry import to_point3d
    from ladybug_rhino.fromgeometry import from_mesh3d, from_vector3d
    from ladybug_rhino.fromobjects import legend_objects, compass_objects
    from ladybug_rhino.text import text_objects
    from ladybug_rhino.grasshopper import all_required_inputs, \
        objectify_output, list_to_data_tree
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


def translate_dome(lb_mesh, lb_compass, graphic, title_txt, values):
    """Translate sky dome geometry into a format suitable for Rhino.

    Args:
        lb_mesh: A ladybug Mesh3D for the dome.
        lb_compass: A ladybug Compass object.
        graphic: A GraphicContainer for the dome.
        title_txt: Text for title of the dome.
        values: A list of radiation values that align with the dome_mesh faces.

    Returns:
        dome_mesh: A Rhino colored mesh for the dome.
        dome_compass: Rhino objects for the dome compass.
        dome_legend:  Rhino objects for the dome legend.
        dome_title: A bake-able title for the dome.
        values: A list of radiation values that align with the dome_mesh faces.
    """
    # translate the dome visualization, including legend and compass
    dome_mesh = from_mesh3d(lb_mesh)
    dome_legend = legend_objects(graphic.legend)
    dome_compass = compass_objects(
        lb_compass, z, None, sky_dome.projection, graphic.legend_parameters.font)

    # construct a title from the metadata
    dome_title = text_objects(title_txt, graphic.lower_title_location,
                              graphic.legend_parameters.text_height,
                              graphic.legend_parameters.font)

    return dome_mesh, dome_compass, dome_legend, dome_title, values


if all_required_inputs(ghenv.Component):
    # set defaults for global variables
    _scale_ = 1 if _scale_ is None else _scale_
    radius = (100 * _scale_) / conversion_to_meters()
    if _center_pt_ is not None:  # process the center point
        center_pt3d = to_point3d(_center_pt_)
        z = center_pt3d.z
    else:
        center_pt3d, z = Point3D(), 0

    # create the SkyDome object
    sky_dome = SkyDome(_sky_mtx, legend_par_, irradiance_,
                       center_pt3d, radius, projection_)

    # output patch patch vectors
    patch_vecs = [from_vector3d(vec) for vec in sky_dome.patch_vectors]

    # create the dome visualization
    if not show_comp_:  # only create the total dome mesh
        mesh, compass, legend, title, mesh_values = translate_dome(*sky_dome.draw())
        patch_values = sky_dome.total_values
    else:  # create domes for total, direct and diffuse
        # loop through the 3 radiation types and produce a dome
        mesh, compass, legend, title, mesh_values = [], [], [], [], []
        rad_types = ('total', 'direct', 'diffuse')
        for dome_i in range(3):
            cent_pt = Point3D(center_pt3d.x + radius * 3 * dome_i,
                              center_pt3d.y, center_pt3d.z)
            dome_mesh, dome_compass, dome_legend, dome_title, dome_values = \
                translate_dome(*sky_dome.draw(rad_types[dome_i], cent_pt))
            mesh.append(dome_mesh)
            compass.extend(dome_compass)
            legend.extend(dome_legend)
            title.append(dome_title)
            mesh_values.append(dome_values)
        rad_data = (sky_dome.total_values, sky_dome.direct_values, sky_dome.diffuse_values)
        patch_values = list_to_data_tree(rad_data)
        mesh_values = list_to_data_tree(mesh_values)

    # output the visualization set
    vis_set = [sky_dome, show_comp_]
    vis_set = objectify_output('VisualizationSet Aruments [SkyDome]', vis_set)



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Calculate hour of the year from month, day, hour, minute.
-

    Args:
        _month_: Integer for month (1-12).
        _day_: Integer for day (1-31).
        _hour_: Integer for hour (0-23).
        _minute_: Integer for minute (0-59).

    Returns:
        hoy: Hour of the year.
        doy: Day of the year.
        date: Human readable date time.
"""

ghenv.Component.Name = 'LB Calculate HOY'
ghenv.Component.NickName = 'HOY'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '1 :: Analyze Data'
ghenv.Component.AdditionalHelpFromDocStrings = '2'

try:
    from ladybug.dt import DateTime
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

month = _month_ or 9
day = _day_ or 21
hour = _hour_ if _hour_ is not None else 12
minute = _minute_ or 0

datetime = DateTime(month, day, hour, minute)
hoy = datetime.hoy
doy = datetime.doy
date = datetime



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Deconstruct a Ladybug DataCollection into a header and values.
-

    Args:
        _data: A Ladybug DataCollection object.
    Returns:
        header: The header of the DataCollection (containing metadata).
        values: The numerical values of the DataCollection.
"""

ghenv.Component.Name = "LB Deconstruct Data"
ghenv.Component.NickName = 'XData'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '1 :: Analyze Data'
ghenv.Component.AdditionalHelpFromDocStrings = '1'

try:
    from ladybug.datacollection import BaseCollection
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    assert isinstance(_data, BaseCollection), \
        '_data must be a Data Collection. Got {}.'.format(type(_data))
    header = _data.header
    values = _data.values


# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Create a set of parameters that define the acceptable conditions of the
Universal Thermal Climate Index (UTCI) comfort model.
-
These parameters can be plugged into any of the components that compute UTCI comfort.
-

    Args:
        _cold_thresh_: Temperature in Celsius below which the UTCI represents
            cold stress. (Default: 9C).
        heat_thresh: Temperature in Celsius above which the UTCI represents
            heat stress. (Default: 26C).

    Returns:
        utci_par: A UTCI comfort parameter object that can be plugged into
            any of the components that compute UTCI thermal comfort.
"""

ghenv.Component.Name = 'LB UTCI Comfort Parameters'
ghenv.Component.NickName = 'UTCIPar'
ghenv.Component.Message = '1.6.1'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '4 :: Extra'
ghenv.Component.AdditionalHelpFromDocStrings = '0'

try:
    from ladybug_comfort.parameter.utci import UTCIParameter
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_comfort:\n\t{}'.format(e))

moderate_cold, moderate_heat = None, None
if _cold_thresh_ and _cold_thresh_ < 0:
    moderate_cold = _cold_thresh_
if _heat_thresh_ and _heat_thresh_ > 28:
    moderate_heat = _heat_thresh_

utci_par = UTCIParameter(
    cold_thresh=_cold_thresh_,
    heat_thresh=_heat_thresh_,
    moderate_cold_thresh=moderate_cold,
    moderate_heat_thresh=moderate_heat
)



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Calculate Mean Radiant Temperature (MRT) as a result of shortwave solar using
horizontal solar components (direct horizontal and diffuse horizontal solar).
-
This component uses the SolarCal model of ASHRAE-55 to estimate the effects
of shortwave solar and a simple sky exposure method to determine longwave
radiant exchange.
-

    Args:
        _location: A Ladybug Location object.
        _longwave_mrt: A single number or an hourly data collection with the long-wave
            mean radiant temperature around the person in degrees C.
            This includes the temperature of the ground and any other
            surfaces between the person and their view to the sky.
            Typically, air temperature is used when such surface
            temperatures are unknown.
        _dir_horiz_rad: Hourly Data Collection with the direct horizontal solar
            irradiance in W/m2.
        _diff_horiz_rad: Hourly Data Collection with diffuse horizontal solar
            irradiance in W/m2.
        fract_body_exp_: A single number between 0 and 1 or a data collection
            representing the fraction of the body exposed to direct sunlight.
            Note that this does not include the body’s self-shading; only the
            shading from surroundings.
            Default is 1 for a person standing in an open area.
        _ground_ref_: A single number between 0 and 1 or a data collection
            that represents the reflectance of the floor. Default is for 0.25
            which is characteristic of outdoor grass or dry bare soil.
        _solar_body_par_: Optional solar body parameters from the "LB Solar Body Parameters"
            object to specify the properties of the human geometry assumed in the
            shortwave MRT calculation. The default assumes average skin/clothing
            absorptivity and a human subject always has their back to the sun
            at a 45-degree angle (SHARP = 135).
        _run: Set to True to run the component.

    Returns:
        report: Reports, errors, warnings, etc.
        erf: Data collection of effective radiant field (ERF) in W/m2.
        dmrt: Data collection of mean radiant temperature delta in C.
        mrt: Data collection of mean radiant temperature in C.
"""

ghenv.Component.Name = 'LB Solar MRT from Solar Components'
ghenv.Component.NickName = 'ComponentSolarMRT'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '1 :: Analyze Data'
ghenv.Component.AdditionalHelpFromDocStrings = '0'

try:
    from ladybug_comfort.collection.solarcal import HorizontalSolarCal
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component) and _run is True:
    solar_mrt_obj = HorizontalSolarCal(_location, _dir_horiz_rad,
                                       _diff_horiz_rad, _longwave_mrt,
                                       fract_body_exp_, _ground_ref_, _solar_body_par_)

    erf = solar_mrt_obj.effective_radiant_field
    dmrt = solar_mrt_obj.mrt_delta
    mrt = solar_mrt_obj.mean_radiant_temperature


# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Visualize the portion of the sky dome that is masked by context geometry or shading
strategies around a given point.
_
Separate meshs will be generated for the portions of the sky dome that are masked
vs visible. The percentage of the sky that is masked by the context geometry and
is visible will also be computed.
-

    Args:
        context_: Rhino Breps and/or Rhino Meshes representing context geometry that
            can block the sky to the center of the sky mask.
        orientation_: A number between 0 and 360 that sets the direction of a vertically-
            oriented surface for which the sky view will be visualized and
            computed. Alternatively, this input can be the words "north",
            "east", "south" or "west." An input here will result in the output
            of an orient_mask, which blocks the portion of the sky that is not
            visible from a vertical surface with this orientation. Furthermore,
            all of the view-related outputs will be computed for a surface with
            the specified orientation (overriding any plane input for the _center_).
        overhang_proj_: A number between 0 and 90 that sets the angle between the _center_
            and the edge of an imagined horizontal overhang projecting past
            the point. Note that this option is only available when there is
            an input for orientation_ above. An input here will result in the
            output of a strategy_mask, which blocks the portion of the sky
            taken up by an overhang with the input projection angle.
        left_fin_proj_: A number between 0 and 180 that sets the angle between the _center_
            and the edge of an imagined vertical fin projecting past the left
            side of the point. Note that this option is only available when
            there is an input for orientation_ above. An input here will result
            in the output of a strategy_mask, which blocks the portion of the
            sky taken up by a vertical fin with the input projection angle.
        right_fin_proj_: A number between 0 and 180 that sets the angle between the _center_
            and the edge of an imagined vertical fin projecting past the right
            side of the point. Note that this option is only available when
            there is an input for orientation_ above. An input here will result
            in the output of a strategy_mask, which blocks the portion of the
            sky taken up by a vertical fin with the input projection angle.
        _density_: An integer that is greater than or equal to 1, which to sets the
            number of times that the sky patches are split. Higher numbers
            input here will ensure a greater accuracy but will also take
            longer to run. A value of 3 should result in sky view factors
            with less than 1% error from the true value. (Default: 1).
        _center_: A point or plane for which the visible portion of the sky will
            be evaluated. If a point is input here, the view-related outputs
            will be indiferent to orientation and the sky_view outut will
            technically be Sky Exposure (or the fraction of the sky hemisphere
            that is visible from the point). If a plane is input here (or an
            orientation_ is connected), the view-related outputs will be
            sensitive to orientation and the sky_view output will be true
            Sky View (or the fraction of the sky visible from a surface in
            a plane). If no value is input here, the center will be a point
            (Sky Exposure) at the Rhino origin (0, 0, 0).
        _scale_: A number to set the scale of the sky mask. The default is 1,
            which corresponds to a radius of 100 meters in the current Rhino
            model's unit system.
        projection_: Optional text for the name of a projection to use from the sky
            dome hemisphere to the 2D plane. If None, a 3D dome will be drawn
            instead of a 2D one. Choose from the following:
                * Orthographic
                * Stereographic

    Returns:
        report: ...
        context_mask: A mesh for the portion of the sky dome masked by the context_
            geometry.
        orient_mask: A mesh for the portion of the sky dome that is not visible from
            a surface is facing a given orientation.
        strategy_mask: A mesh of the portion of the sky dome masked by the overhang,
            left fin, and right fin projections.
        sky_mask: A mesh of the portion of the sky dome visible by the _center_
            through the strategies and context_ geometry.
        context_view: The percentage of the sky dome masked by the context_ geometry.
        orient_view: The percentage of the sky dome that is not visible from a
            surface is facing a given orientation.
        strategy_view: The percentage of the sky dome viewed by the overhang, left
            fin, and right fin projections.
        sky_view: The percentage of the sky dome visible by the _center_ through
            the strategies and context_ geometry.
"""

ghenv.Component.Name = 'LB Sky Mask'
ghenv.Component.NickName = 'SyMask'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '3 :: Analyze Geometry'
ghenv.Component.AdditionalHelpFromDocStrings = '3'

import math

try:
    from ladybug_geometry.geometry2d.pointvector import Point2D
    from ladybug_geometry.geometry3d.pointvector import Point3D, Vector3D
    from ladybug_geometry.geometry3d.mesh import Mesh3D
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_geometry:\n\t{}'.format(e))

try:
    from ladybug.viewsphere import view_sphere
    from ladybug.compass import Compass
    from ladybug.color import Color
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.config import conversion_to_meters
    from ladybug_rhino.togeometry import to_point3d, to_plane
    from ladybug_rhino.fromgeometry import from_mesh3d, from_point3d, from_vector3d
    from ladybug_rhino.intersect import join_geometry_to_mesh, intersect_mesh_rays
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


def apply_mask_to_sky(sky_pattern, mask_pattern):
    """Apply a pattern of a mask to a pattern of visible sky."""
    for i, val in enumerate(mask_pattern):
        if val:
            sky_pattern[i] = False


def apply_mask_to_base_mask(base_pattern, mask_pattern, ignore_pattern):
    """Apply a pattern of a mask to a base pattern with the option to ignore elements."""
    for i, (val, ig_val) in enumerate(zip(mask_pattern, ignore_pattern)):
        if val and not ig_val:
            base_pattern[i] = True


def mask_mesh_from_pattern(base_mask, mask_pattern, color):
    """Get a Rhino mesh of a mask from a pattern aligned to the faces of a base mesh."""
    try:
        mask_mesh = base_mask.remove_faces_only(mask_pattern)
    except AssertionError:  # all mesh faces have been removed
        return None
    mask_mesh.colors = [color] * len(mask_mesh.faces)
    return from_mesh3d(mask_mesh)



# process the inputs and set defaults for global variables
_scale_ = 1 if _scale_ is None else _scale_
radius = (100 * _scale_) / conversion_to_meters()
if _center_ is not None:  # process the center point into a Point2D
    try:  # assume that it is a point
        center_pt3d, direction = to_point3d(_center_), None
    except AttributeError:
        plane, is_orient = to_plane(_center_), True
        center_pt3d, direction = plane.o, plane.n
else:
    center_pt3d, direction = Point3D(), None
az_count, alt_count = 72, 18
if _density_:
    az_count = az_count * _density_
    alt_count = alt_count * _density_
if orientation_ is not None:  # process the orientation to a number
    ori_dict = {'north': 0, 'east': 90, 'south': 180, 'west': 270}
    try:  # first check if it's text for the direction
        orientation_ = ori_dict[orientation_.lower()]
    except KeyError:  # it's a number for the orientation
        orientation_ = float(orientation_)
    direction = Vector3D(0, 1, 0).rotate_xy(-math.radians(orientation_))


# create the dome mesh of the sky and position/project it correctly
sky_mask, view_vecs = view_sphere.dome_radial_patches(az_count, alt_count)
sky_mask = sky_mask.scale(radius)
if center_pt3d != Point3D():
    m_vec = Vector3D(center_pt3d.x, center_pt3d.y, center_pt3d.z)
    sky_mask = sky_mask.move(m_vec)
if projection_ is not None:
    if projection_.title() == 'Orthographic':
        pts = (Compass.point3d_to_orthographic(pt) for pt in sky_mask.vertices)
    elif projection_.title() == 'Stereographic':
        pts = (Compass.point3d_to_stereographic(pt, radius, center_pt3d)
               for pt in sky_mask.vertices)
    else:
        raise ValueError(
            'Projection type "{}" is not recognized.'.format(projection_))
    pts3d = tuple(Point3D(pt.x, pt.y, center_pt3d.z) for pt in pts)
    sky_mask = Mesh3D(pts3d, sky_mask.faces)
sky_pattern = [True] * len(view_vecs)  # pattern to be adjusted by the various masks


# account for the orientation and any of the projection strategies
orient_pattern, strategy_pattern = None, None
if direction is not None:
    orient_pattern, dir_angles = view_sphere.orientation_pattern(direction, view_vecs)
    apply_mask_to_sky(sky_pattern, orient_pattern)
    if overhang_proj_ or left_fin_proj_ or right_fin_proj_:
        strategy_pattern = [False] * len(view_vecs)
        if overhang_proj_:
            over_pattern = view_sphere.overhang_pattern(direction, overhang_proj_, view_vecs)
            apply_mask_to_base_mask(strategy_pattern, over_pattern, orient_pattern)
            apply_mask_to_sky(sky_pattern, over_pattern)
        if left_fin_proj_ or right_fin_proj_:
            f_pattern = view_sphere.fin_pattern(direction, left_fin_proj_, right_fin_proj_, view_vecs)
            apply_mask_to_base_mask(strategy_pattern, f_pattern, orient_pattern)
            apply_mask_to_sky(sky_pattern, f_pattern)


# account for any input context
context_pattern = None
if len(context_) != 0:
    shade_mesh = join_geometry_to_mesh(context_)  # mesh the context
    points = [from_point3d(center_pt3d)]
    view_vecs = [from_vector3d(pt) for pt in view_vecs]
    int_matrix, angles = intersect_mesh_rays(shade_mesh, points, view_vecs)
    context_pattern = [val == 0 for val in int_matrix[0]]
    apply_mask_to_sky(sky_pattern, context_pattern)


# get the weights for each patch to be used in view factor calculation
weights = view_sphere.dome_radial_patch_weights(az_count, alt_count)
if direction is not None:
    weights = [wgt * abs(math.cos(ang)) * 2 for wgt, ang in zip(weights, dir_angles)]


# create meshes for the masks and compute any necessary view factors
gray, black = Color(230, 230, 230), Color(0, 0, 0)
context_view, orient_view, strategy_view = 0, 0, 0
if context_pattern is not None:
    context_mask = mask_mesh_from_pattern(sky_mask, context_pattern, black)
    context_view = sum(wgt for wgt, is_viz in zip(weights, context_pattern) if is_viz)
if orient_pattern is not None:
    orient_mask = mask_mesh_from_pattern(sky_mask, orient_pattern, black)
    orient_view = sum(wgt for wgt, is_viz in zip(weights, orient_pattern) if is_viz)
if strategy_pattern is not None:
    strategy_mask = mask_mesh_from_pattern(sky_mask, strategy_pattern, black)
    strategy_view = sum(wgt for wgt, is_viz in zip(weights, strategy_pattern) if is_viz)
sky_mask = mask_mesh_from_pattern(sky_mask, sky_pattern, gray)
sky_view = sum(wgt for wgt, is_viz in zip(weights, sky_pattern) if is_viz)



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Generate a solar envelope boundary for a given geometry, set of sun vectors, and
context (obstacle) geometry.
_
Solar collection envelopes show the height above which one will have solar access
to certain sun positions on a given site.
_
Solar rights envelopes illustrate the volume in which one can build while ensuring
that a new development does not shade the surrounding properties for certain sun
positions.
-

    Args:
        _geometry: Rhino Breps and/or Rhino Meshes for which the solar envelope will
            be computed. If Breps are input, they will be subdivided using
            the _grid_size to yeild individual points at which analysis will
            occur. If a Mesh is input, the analysis will be performed for each
            vertex of the mesh instead of subdividing it.
        _obstacles: A list of horizontal planar Breps or curves indicating the tops (in the
            case of solar collection) or bottoms (in the case of solar rights)
            of context geometries. Being above a solar collection boundary
            ensures these top surfaces don't block the sun vectors to ones
            position. Being below a solar rights boundary ensures these bottom
            surfaces are protected from shade.
        _vectors: Sun vectors from the "LB SunPath" component, which determine the
            times of the year when sun should be accessible.
        _grid_size: A positive number in Rhino model units for the size of grid
            cells at which the input _geometry will be subdivided for envelope
            analysis. The smaller the grid size, the higher the resolution of
            the analysis and the longer the calculation will take.  So it is
            recommended that one start with a large value here and decrease
            the value as needed. The default will be a relativel coarse
            auto-calculated from the bounding box around the _geometry.
        _height_limit_: A positive number for the minimum distance below (for collections)
            or maximum distance above (for rights) the average _geometry height
            that the envelope points can be. This is used when there are no
            vectors blocked for a given point. (Default: 100 meters).
        solar_rights_: Set to True to compute a solar rights boundary and False to compute
            a solar collection boundary. Solar rights boundaries represent the
            boundary below which one can build without shading the surrounding
            obstacles from any of the _vectors. Solar collection boundaries
            represent the boundary above which one will have direct solar
            access to all of the input _vectors. (Default: False).
        _run: Set to "True" to run the component and get a solar envelope.

    Returns:
        report: ...
        points: The grid of points above the test _geometry representing the height to
            which the solar envelope boundary reaches.
        mesh: A mesh representing the solar envelope. For solar collections (the default),
            this represents the boundary above which the one will have direct
            solar access to all of the input _vectors. For solar rights envelopes,
            this represents the boundary below which one can build without shading
            the surrounding obstacles from any of the _vectors.
"""

ghenv.Component.Name = 'LB Solar Envelope'
ghenv.Component.NickName = 'SolarEnvelope'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '3 :: Analyze Geometry'
ghenv.Component.AdditionalHelpFromDocStrings = '3'

try:
    from ladybug.solarenvelope import SolarEnvelope
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.config import conversion_to_meters
    from ladybug_rhino.togeometry import to_joined_gridded_mesh3d, to_face3d, \
        to_vector3d
    from ladybug_rhino.fromgeometry import from_mesh3d, from_point3d
    from ladybug_rhino.grasshopper import all_required_inputs, hide_output
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component) and _run:
    # set the default offset distance
    _height_limit_ = _height_limit_ if _height_limit_ is not None \
        else 100 / conversion_to_meters()

    # convert geometry, objstacles, and vectors to ladybug_geomtery
    study_mesh = to_joined_gridded_mesh3d(_geometry, _grid_size)
    obstacle_faces = [g for geo in _obstacles for g in to_face3d(geo)]
    sun_vectors = [to_vector3d(vec) for vec in _vectors]

    # compute the solar envelope
    solar_obj = SolarEnvelope(
        study_mesh, obstacle_faces, sun_vectors, _height_limit_, solar_rights_)
    lb_mesh = solar_obj.envelope_mesh()
    mesh = from_mesh3d(lb_mesh)
    points = [from_point3d(pt) for pt in lb_mesh.vertices]
    hide_output(ghenv.Component, 1)



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Convert a Ladybug DataCollection into its Ladybug Legacy format.
-
Note that this component is intended to be temporary as people transition from
Ladybug Legacy to Ladybug[+].
-

    Args:
        _data: A Ladybug DataCollection object.
    
    Returns:
        data: A Ladybug Legacy list with meatadata and values.
"""

ghenv.Component.Name = "LB Data to Legacy"
ghenv.Component.NickName = 'ToLegacy'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '0 :: Import'
ghenv.Component.AdditionalHelpFromDocStrings = '0'

try:
    from ladybug.datacollection import HourlyContinuousCollection, \
        MonthlyCollection, DailyCollection, MonthlyPerHourCollection
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))

# base header list
head = ['key:location/dataType/units/frequency/startsAt/endsAt',
        'Unkown Location']


if all_required_inputs(ghenv.Component):
    # add the location, data type, and units
    meta_data = _data.header.metadata
    if 'city' in meta_data:
        head[1] = meta_data['city']
    if 'type' in meta_data:
        if 'Zone' in meta_data:
            head.append('{} for {}'.format(meta_data['type'], meta_data['Zone']))
        elif 'System' in meta_data:
            head.append('{} for {}'.format(meta_data['type'], meta_data['System']))
        else:
            head.append(meta_data['type'])
    else:
        head.append(str(_data.header.data_type))
    head.append(_data.header.unit)

    # add the time interval
    a_per = _data.header.analysis_period
    if isinstance(_data, HourlyContinuousCollection):
        if a_per.timestep == 1:
            head.append('Hourly')
        else:
            head.append('Timestep')
    elif isinstance(_data, MonthlyCollection):
        head.append('Monthly')
    elif isinstance(_data, DailyCollection):
        head.append('Daily')
    elif isinstance(_data, MonthlyPerHourCollection):
        head.append('Monthly-> averaged for each hour')
    else:
        raise TypeError(
            '_data must be a Data Collection. Got {}.'.format(type(_data)))

    # add the analysis period
    head.append((a_per.st_month, a_per.st_day, a_per.st_hour + 1))
    head.append((a_per.end_month, a_per.end_day, a_per.end_hour + 1))

    # return the data
    data = head + list(_data.values)



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Customize the properties of a screen-oreinted 2D legend displaying with the
"LB Preview VisualizationSet" component.
-

    Args:
        origin_x_: An integer in pixels to note the X coordinate of the base point from
            where the 2D legend will be generated (assuming an origin in the
            upper-left corner of the screen with higher positive values of
            X moving to the right). Alternatively, this can be a text string
            ending in a % sign to denote the percentage of the screen where
            the X coordinate exists (eg. 5%). The default is set to make the
            legend clearly visible in the upper-left corner of the
            screen (10 pixels).
        origin_y_: An integer in pixels to note the Y coordinate of the base point from
            where the legend will be generated (assuming an origin in the
            upper-left corner of the screen with higher positive values of
            Y moving downward). Alternatively, this can be a text string
            ending in a % sign to denote the percentage of the screen where
            the X coordinate exists (eg. 5%). The default is set to make the
            legend clearly visible in the upper-left corner of the
            screen (50 pixels).
        seg_height_: A integer in pixels to note the height for each of the legend segments.
            Alternatively, this can be a text string ending in a % sign to
            denote the percentage of the screen (eg. 5%). The default is set
            to make most legends readable on standard resolution
            screens (25px for horizontal and 36px for vertical).
        seg_width_: An integer in pixels to set the width of each of the legend segments.
            Alternatively, this can be a text string ending in a % sign to
            denote the percentage of the screen (eg. 5%). The default is set
            to make most legends readable on standard resolution
            screens (36px for horizontal and 25px for vertical).
        text_height_: An integer in pixels to set the height for the legend text.
            Alternatively, this can be a text string ending in a % sign to
            denote the percentage of the screen (eg. 2%).

    Returns:
        leg_par: A legend parameter object that can be plugged into the "LB Preview
            VisualizationSet" component to specify the properties of a
            screen-oriented legend.
"""

ghenv.Component.Name = 'LB Legend 2D Parameters'
ghenv.Component.NickName = 'Legend2D'
ghenv.Component.Message = '1.6.1'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '4 :: Extra'
ghenv.Component.AdditionalHelpFromDocStrings = '0'


try:
    from ladybug.legend import Legend2DParameters
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))


def parse_dim_text(dim_text):
    """Parse text representing a dimension into an input for legend parameters."""
    try:
        px_txt = int(dim_text)
        return '{}px'.format(px_txt)
    except ValueError:
        return dim_text


# parse all of the inputs
origin_x_ = parse_dim_text(origin_x_) if origin_x_ is not None else None
origin_y_ = parse_dim_text(origin_y_) if origin_y_ is not None else None
seg_height_ = parse_dim_text(seg_height_) if seg_height_ is not None else None
seg_width_ = parse_dim_text(seg_width_) if seg_width_ is not None else None
text_height_ = parse_dim_text(text_height_) if text_height_ is not None else None

# make the 2D legend parameters
leg_par2d = Legend2DParameters(origin_x_, origin_y_, seg_height_, seg_width_, text_height_)



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Create a compass sign that indicates the direction of North in the Rhino scene.
-

    Args:
        _north_: A number between -360 and 360 for the counterclockwise difference
            between the North and the positive Y-axis in degrees. Counterclockwise
            means "90 is West and 270 is East". This can also be Vector for the
            direction to North. (Default: 0)
        _center_: A point for the center position of the compass in the Rhino
            scene. (Default: (0, 0, 0) aka. the Rhino scene origin).
        _scale_: A number to set the scale of the compass. The default is 1,
            which corresponds to a radius of 10 meters in the current Rhino
            model's unit system.

    Returns:
        compass: A set of circles, lines and text objects that mark the cardinal
            directions in the Rhino scene.
"""

ghenv.Component.Name = 'LB Compass'
ghenv.Component.NickName = 'Compass'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '4 :: Extra'
ghenv.Component.AdditionalHelpFromDocStrings = '6'

import math

try:
    from ladybug_geometry.geometry2d.pointvector import Vector2D, Point2D
    from ladybug_geometry.geometry3d.pointvector import Point3D, Vector3D
    from ladybug_geometry.geometry3d.plane import Plane
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_geometry:\n\t{}'.format(e))

try:
    from ladybug.compass import Compass
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.config import conversion_to_meters
    from ladybug_rhino.togeometry import to_vector2d, to_point2d, to_point3d
    from ladybug_rhino.fromgeometry import from_arc2d, from_linesegment2d
    from ladybug_rhino.text import text_objects
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


def translate_compass(compass, z=0, font='Arial'):
    """Translate a Ladybug Compass object into Grasshopper geometry.

    Args:
        compass: A Ladybug Compass object to be converted to Rhino geometry.
        z: A number for the Z-coordinate to be used in translation. (Default: 0)
        font: Optional text for the font to be used in creating the text.
            (Default: 'Arial')

    Returns:
        A list of Rhino geometries in the following order.
        -   all_boundary_circles -- Three Circle objects for the compass boundary.
        -   major_azimuth_ticks -- Line objects for the major azimuth labels.
        -   major_azimuth_text -- Bake-able text objects for the major azimuth labels.
     """
    # set default variables based on the compass properties
    maj_txt = compass.radius / 2.5
    xaxis = Vector3D(1, 0, 0).rotate_xy(math.radians(compass.north_angle))
    result = []  # list to hold all of the returned objects

    # create the boundary circles
    for circle in compass.all_boundary_circles:
        result.append(from_arc2d(circle, z))

    # generate the labels and tick marks for the azimuths
    for line in compass.major_azimuth_ticks:
        result.append(from_linesegment2d(line, z))
    for txt, pt in zip(compass.MAJOR_TEXT, compass.major_azimuth_points):
        result.append(text_objects(
            txt, Plane(o=Point3D(pt.x, pt.y, z), x=xaxis), maj_txt, font, 1, 3))
    return result


# set defaults for all of the
if _north_ is not None:  # process the _north_
    try:
        _north_ = math.degrees(to_vector2d(_north_).angle_clockwise(Vector2D(0, 1)))
    except AttributeError:  # north angle instead of vector
        _north_ = float(_north_)
else:
    _north_ = 0
if _center_ is not None:  # process the center point into a Point2D
    center_pt, z = to_point2d(_center_), to_point3d(_center_).z
else:
    center_pt, z = Point2D(), 0
_scale_ = 1 if _scale_ is None else _scale_ # process the scale into a radius
radius = (10 * _scale_) / conversion_to_meters()

# create the compass
compass = translate_compass(Compass(radius, center_pt, _north_, 1), z)



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Use this component to access a library of typical gradients useful throughout Ladybug. 
The output from this component should be plugged into the colors_ input of the
"Legend Parameters" component.
-
For an image of each of the gardients in the library, check here:
https://github.com/ladybug-tools/lbt-grasshopper/blob/master/gradients.png
-

    Args:
        _index_: An index refering to one of the following possible gradients:
            0 - Original Ladybug
            1 - Nuanced Ladybug
            2 - Multi-colored Ladybug
            3 - Ecotect
            4 - View Study
            5 - Shadow Study
            6 - Glare Study
            7 - Annual Comfort
            8 - Thermal Comfort
            9 - Peak Load Balance
            10 - Heat Sensation
            11 - Cold Sensation
            12 - Benefit/Harm
            13 - Harm
            14 - Benefit
            15 - Shade Benefit/Harm
            16 - Shade Harm
            17 - Shade Benefit
            18 - Energy Balance
            19 - Energy Balance w/ Storage
            20 - THERM
            21 - Cloud Cover
            22 - Black to White
            23 - Blue, Green, Red
            24 - Multicolored 2
            25 - Multicolored 3
            26 - OpenStudio Palette

    Returns:
        colors: A series of colors to be plugged into the "LB Legend Parameters"
            component.
"""

ghenv.Component.Name = 'LB Color Range'
ghenv.Component.NickName = 'ColRange'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '4 :: Extra'
ghenv.Component.AdditionalHelpFromDocStrings = '1'

try:
    from ladybug.color import Colorset
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.color import color_to_color
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


_index_ = _index_ or 0
cs = Colorset()
colors = [color_to_color(col) for col in cs[_index_]]



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Estimate levels of clothing using a temperature value or data collection of
temperatures to which a human subject is adapting (typically the outdoor
air temperature).
_
This resulting clothing values can be plugged into the _clothing_ input of the
"LB PMV Comfort" component or the "LB PET Comfort" component. They can also
be used in thermal mapping recipes.
_
By default, this function derives clothing levels using a model developed by
Schiavon, Stefano based on outdoor air temperature, which is implemented in the
CBE comfort tool (https://comfort.cbe.berkeley.edu/).
_
The version of the model implemented here allows changing of the maximum and minimum
clothing levels, which the Schiavon model sets at 1 and 0.46 respectively, and the
temperatures at which these clothing levels occur, which the Schiavon model sets
at -5 C and 26 C respectively.
-

    Args:
        _temperature: A data collection or single number representing the temperature to
            which the human subject adapts their clothing. This is typically the
            dry bulb temperature obtained from the "LB Import EPW" component.
        _period_: An optional analysis period to select a subset of the whole outdoor
            temperature data collection for which clothing levels will be
            computed. By default the whole analysis period of the _temperature
            will be used.
        _max_clo_: A number for the maximum clo value that the human subject wears
            on the coldest days. (Default: 1 clo, per the original Schiavon
            clothing function).
        _max_clo_temp_: A number for the temperature below which the _max_clo_ value
            is applied (in Celsius). (Default: -5 C, per the original
            Schiavon clothing function with outdoor temperature).
        _min_clo_: A number for the minimum clo value that the human subject wears
            wears on the hotest days. (Default: 0.46 clo,
            per the original Schiavon clothing function).
        _min_clo_temp_: A number for the temperature above which the _min_clo_ value
            is applied (in Celsius). (Default: 26 C, per the original
            Schiavon clothing function).

    Returns:
        report: Reports, errors, warnings, etc.
        clo: A single number or data collection of numbers representing the clothing
            that would be worn (in clo). Note that, if you have hooked up an
            hourly continuous data collection, the clothing levels will change
            on a 12-hour basis to simulate the typical cycle on which a human
            changes their clothing.

"""

ghenv.Component.Name = "LB Clothing by Temperature"
ghenv.Component.NickName = 'CloByTemp'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '1 :: Analyze Data'
ghenv.Component.AdditionalHelpFromDocStrings = '0'

try:
    from ladybug.datacollection import BaseCollection, HourlyContinuousCollection
    from ladybug.datatype.rvalue import ClothingInsulation
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_comfort.clo import schiavon_clo
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # set default values
    _max_clo_ = _max_clo_ if _max_clo_ is not None else 1.0
    _max_clo_temp_ = _max_clo_temp_ if _max_clo_temp_ is not None else -5
    _min_clo_ = _min_clo_ if _min_clo_ is not None else 0.46
    _min_clo_temp_ = _min_clo_temp_ if _min_clo_temp_ is not None else 26

    # if the temperature is hourly continuous, simplify the values
    if isinstance(_temperature, HourlyContinuousCollection):
        date_times, temps = _temperature.datetimes, _temperature.values
        last_time = date_times[0].sub_hour(18)  # clothing determined at 6 AM
        last_val = temps[0]
        new_vals = []
        for v, dt in zip(temps, date_times):
            time_diff = dt - last_time
            if time_diff.days >= 1:
                last_time, last_val = dt, v
            new_vals.append(last_val)
        _temperature = _temperature.duplicate()
        _temperature.values = new_vals

    # apply the analysis period if requests
    if period_ is not None and isinstance(_temperature, BaseCollection):
        _temperature = _temperature.filter_by_analysis_period(period_)

    clo = HourlyContinuousCollection.compute_function_aligned(
        schiavon_clo, [_temperature, _max_clo_, _max_clo_temp_, _min_clo_, _min_clo_temp_],
        ClothingInsulation(), 'clo')



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Set the sun in the Rhino scene to correspond to a given location and date/time.
_
This can be help coordinate Rhino visualizations with Ladybug analyses.
-

    Args:
        _location: A ladybug Location that has been output from the "LB Import EPW"
            component or the "LB Construct Location" component.
        hoys_: A number between 0 and 8760 that represents the hour of the year at
            which to set the sun position. The "LB Calculate HOY" component
            can output this number given a month, day and hour. The "LB
            Analysis Period" component can output a list of HOYs within a
            certain hour or date range.
        north_: A number between -360 and 360 for the counterclockwise
            difference between the North and the positive Y-axis in degrees.
            90 is West and 270 is East. This can also be Vector for the
            direction to North. (Default: 0)
        _run: Set to True to run the component set the Rhino Sun.

    Returns:
        report: Reports, errors, warnings, etc.
"""

ghenv.Component.Name = 'LB Set Rhino Sun'
ghenv.Component.NickName = 'RhinoSun'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '3 :: Analyze Geometry'
ghenv.Component.AdditionalHelpFromDocStrings = '3'

import math

try:
    from ladybug_geometry.geometry2d.pointvector import Vector2D
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_geometry:\n\t{}'.format(e))

try:
    from ladybug_rhino.togeometry import to_vector2d
    from ladybug_rhino.light import set_sun, disable_sun
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))

if all_required_inputs(ghenv.Component) and _run:
    if north_ is not None:  # process the north_
        try:
            north_ = math.degrees(
                to_vector2d(north_).angle_clockwise(Vector2D(0, 1)))
        except AttributeError:  # north angle instead of vector
            north_ = float(north_)
    else:
        north_ = 0
    set_sun(_location, _hoy, north_)
else:
    disable_sun()



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Construct location from latitude, lognitude, and time zone data.
-

    Args:
        _name_: A name for the location you are constructing. For example,
            "Steventon Island, Antarctica". (Default: "-")
        _latitude_: Location latitude between -90 and 90 (Default: 0).
        _longitude_: Location longitude between -180 (west) and 180 (east) (Default: 0).
        _time_zone_: Time zone between -12 hours (west) and 12 hours (east). If None,
            the time zone will be an estimated integer value derived from the
            longitude in accordance with solar time (Default: None).
        _elevation_: A number for elevation of the location in meters. (Default: 0).
    
    Returns:
        location: Location data (use this output to construct the sun path).
"""

ghenv.Component.Name = 'LB Construct Location'
ghenv.Component.NickName = 'ConstrLoc'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '0 :: Import'
ghenv.Component.AdditionalHelpFromDocStrings = '3'

try:
    from ladybug.location import Location
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))

location = Location(_name_, '-', '-', _latitude_, _longitude_, _time_zone_, _elevation_)



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Calculate Universal Thermal Climate Index (UTCI).
-
UTCI is a thermal comfort model strictly for the outdoors. It is an international
standard for outdoor temperature sensation (aka. "feels-like" temperature) and
is one of the most common of such "feels-like" temperatures used by
meteorologists. UTCI that attempts to satisfy the following requirements:
_
1) Thermo-physiological significance in the whole range of heat exchange conditions.
2) Valid in all climates, seasons, and scales.
3) Useful for key applications in human biometeorology.
_
While UTCI is designed to be valid in all climates and seasons, it assumes
that human subjects are walking (with a metabolic rate around 2.4 met) and
that they naturally adapt their clothing with the outdoor temperature.
For outdoor situations that do not fit these criteria, the Physiological
Equivalent Temperature (PET) model is recommended.
-

    Args:
        _air_temp: Data Collection or individual value for air temperature in C.
        _mrt_: Data Collection or individual value for mean radiant temperature
            (MRT) in C. Default is the same as the air_temp.
        _rel_humid: Data Collection or individual value for relative humidity in %.
            Note that percent values are between 0 and 100.
        _wind_vel_: Data Collection or individual value for meteoroligical wind velocity
            at 10 m above ground level in m/s. Note that this meteorological
            velocity at 10 m is simply 1.5 times the speed felt at ground
            level in the original Fiala model used to create the UTCI model.
            Therefore, multiplying air speed values at the height of the
            human subject by 1.5 will make them a suitable input for this
            component. Default is a low speed of 0.5 m/s, which is the lowest
            input speed that is recommended for the UTCI model.
        utci_par_: Optional comfort parameters from the "LB UTCI Comfort Parameters"
            component to specify the temperatures (in Celcius) that are
            considered acceptable/comfortable. The default will assume a that 
            the comfort range is between 9C and 26C.
        _run: Set to True to run the component.

    Returns:
        report: Reports, errors, warnings, etc.
        utci: Universal Thermal Climate Index (UTCI) in Celcius.
        comfort: Integers noting whether the input conditions result in no
            thermal stress.
            .
            Values are one of the following:
                * 0 = thermal stress
                * 1 = no thremal stress
        condition: Integers noting the thermal status of a subject.
            .
            Values are one of the following:
                * -1 = cold
                *  0 = netural
                * +1 = hot
        category: Integers noting the category that the UTCI conditions fall
            under on an 11-point scale.
            .
            Values are one of the following:
                * -5 = Extreme Cold Stress       (UTCI < -40)
                * -4 = Very Strong Cold Stress   (-40 <= UTCI < -27)
                * -3 = Strong Cold Stress        (-27 <= UTCI < -13)
                * -2 = Moderate Cold Stress      (-12 <= UTCI < 0)
                * -1 = Slight Cold Stress        (0 <= UTCI < 9)
                *  0 = No Thermal Stress         (9 <= UTCI < 26)
                * +1 = Slight Heat Stress        (26 <= UTCI < 28)
                * +2 = Moderate Heat Stress      (28 <= UTCI < 32)
                * +3 = Strong Heat Stress        (32 <= UTCI < 38)
                * +4 = Very Strong Heat Stress   (38 <= UTCI < 46)
                * +5 = Extreme Heat Stress       (46 < UTCI)
        comf_obj: A Python object containing all inputs and results of the
            analysis.  This can be plugged into components like the "Comfort
            Statistics" component to get further information.
"""

ghenv.Component.Name = 'LB UTCI Comfort'
ghenv.Component.NickName = 'UTCI'
ghenv.Component.Message = '1.6.1'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '1 :: Analyze Data'
ghenv.Component.AdditionalHelpFromDocStrings = '5'

try:
    from ladybug.datatype.temperature import Temperature
    from ladybug.datacollection import BaseCollection
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_comfort.collection.utci import UTCI
    from ladybug_comfort.parameter.utci import UTCIParameter
    from ladybug_comfort.utci import universal_thermal_climate_index
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_comfort:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


def extract_collections(input_list):
    """Process inputs into collections and floats."""
    defaults = [None, _air_temp, 0.5, None]
    data_colls = []
    for i, input in enumerate(input_list):
        if input is None:
            input_list[i] = defaults[i]
        elif isinstance(input, BaseCollection):
            data_colls.append(input)
        else:
            try:
                input_list[i] = float(input)
            except ValueError as e:
                raise TypeError('input {} is not valid. Expected float or '
                                'DataCollection. Got {}'.format(input, type(input)))
    return input_list, data_colls

if all_required_inputs(ghenv.Component) and _run is True:
    # Process inputs and assign defaults.
    input_list = [_air_temp, _mrt_, _wind_vel_, _rel_humid]
    input, data_colls = extract_collections(input_list)
    utci_par = utci_par_ or UTCIParameter()

    if data_colls == []:
        # The inputs are all individual values.
        utci = universal_thermal_climate_index(
            input[0], float(input[1]), input[2], input[3])
        utci_par = UTCIParameter()
        comfort = utci_par.is_comfortable(utci)
        condition = utci_par.thermal_condition(utci)
        category = utci_par.thermal_condition_eleven_point(utci)
    else:
        # The inputs include Data Collections.
        if not isinstance(_air_temp, BaseCollection):
            _air_temp = data_colls[0].get_aligned_collection(
                float(_air_temp), Temperature(), 'C')

        comf_obj = UTCI(_air_temp, _rel_humid, _mrt_, _wind_vel_, utci_par)
        utci = comf_obj.universal_thermal_climate_index
        comfort = comf_obj.is_comfortable
        condition = comf_obj.thermal_condition
        category = comf_obj.thermal_condition_eleven_point


# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Calculate Mean Radiant Temperature (MRT) as a result of outdoor shortwave
solar shining directly onto people as well as longwave radiant exchange
with the sky.
-
This component uses the SolarCal model of ASHRAE-55 to estimate the effects
of shortwave solar and a simple sky exposure method to determine longwave
radiant exchange.
-

    Args:
        _location: A Ladybug Location object, used to determine the altitude and
            azimuth of the sun.
        _longwave_mrt: A single number or an hourly data collection with the long-wave
            mean radiant temperature around the person in degrees C.
            This includes the temperature of the ground and any other
            surfaces between the person and their view to the sky.
            Typically, indoor air temperature is used when such surface
            temperatures are unknown.
        _dir_norm_rad: Hourly Data Collection with the direct normal solar
            irradiance in W/m2.
        _diff_horiz_rad: Hourly Data Collection with diffuse horizontal solar
            irradiance in W/m2.
        fract_body_exp_: A single number between 0 and 1 or a data collection for
            the fraction of the body exposed to direct sunlight. The "LB
            Human to Sky Relationship" component can be used to estimate this
            input for a given set of context geometry and position of the
            human. Note that this parameter does NOT include the body’s
            self-shading. It only includes the shading from furniture and
            surroundings. (Default: 1 for an area surrounded by glass).
        sky_exposure_: A single number between 0 and 1 or a data collection representing
            the fraction of the sky vault in the human subject’s view. The "LB
            Human to Sky Relationship" component can be used to estimate this
            input for a given set of context geometry and position of the
            human. (Default: 0.5 for a person next to an all glass facade).
        _ground_ref_: A single number between 0 and 1 or a data collection
            that represents the reflectance of the floor. Default is for 0.25
            which is characteristic of concrete.
        _window_trans_: A Data Collection or number between 0 and 1 that represents the
                broadband solar transmittance of the window through which the
                sun is coming. Such values tend to be slightly less than the
                SHGC. Values might be as low as 0.2 and could be as high as 0.85
                for a single pane of glass. Default is 0.4 assuming a double pane
                window with a relatively mild low-e coating.
        _solar_body_par_: Optional solar body parameters from the "LB Solar Body Parameters"
            object to specify the properties of the human geometry assumed in the
            shortwave MRT calculation. The default assumes average skin/clothing
            absorptivity and a human subject always has their back to the sun
            at a 45-degree angle (SHARP = 135).
        _run: Set to True to run the component.

    Returns:
        report: Reports, errors, warnings, etc.
        erf: Data collection of effective radiant field (ERF) in W/m2.
        dmrt: Data collection of mean radiant temperature delta in C.
        mrt: Data collection of mean radiant temperature in C.
"""

ghenv.Component.Name = 'LB Indoor Solar MRT'
ghenv.Component.NickName = 'IndoorSolarMRT'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '1 :: Analyze Data'
ghenv.Component.AdditionalHelpFromDocStrings = '6'

try:
    from ladybug_comfort.collection.solarcal import IndoorSolarCal
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component) and _run is True:
    sky_exposure_ = 0.5 if sky_exposure_ is None else sky_exposure_
    solar_mrt_obj = IndoorSolarCal(_location, _dir_norm_rad,
                                   _diff_horiz_rad, _longwave_mrt,
                                   fract_body_exp_, sky_exposure_,
                                   _ground_ref_, _window_trans_, _solar_body_par_)

    erf = solar_mrt_obj.effective_radiant_field
    dmrt = solar_mrt_obj.mrt_delta
    mrt = solar_mrt_obj.mean_radiant_temperature


# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Calculate Physiological Equivalent Temperature (PET).
_
PET uses the Munich Energy Balance Model (MEMI), which is arguably the most
detailed 3-node human energy balance model in common use today. It can account
for various physiological features of the human subject, including age, sex,
height, and body mass, making it one of the only models that is suitable for
forecasting the thermal experience of a specific individual. This also makes
it one of the better models for estimating core body temperature and whether
a given set of conditions is likely to induce hypothermia or hyperthermia in
a specific individual.
-

    Args:
        _air_temp: Data Collection or individual value for air temperature in C.
        _mrt_: Data Collection or individual value of mean radiant temperature
            (MRT) in C. Default is the same as the air_temp.
        _rel_humid: Data Collection or individual value for relative humidity in %.
            Note that percent values are between 0 and 100.
        _air_speed_: Data Collection or individual value for air speed in m/s.
            Default is a very low speed of 0.1 m/s, which is typical of the
            room air speeds induced by HVAC systems.
        _pressure_: A value or data collection representing atmospheric pressure [Pa]
            Default is to use air pressure at sea level (101,325 Pa).
        _met_rate_: Data Collection or individual value of metabolic rate in met.
            Note that the original PET model requires that the activity of
            the human subject be accounted for as additional Watts above
            the basal metabolism, which is often difficult to estimate. In
            order to accept an input in [met], it is assumed that 1 met refers
            to Resting Metabolic Rate (RMR) and this is 1.17 times the male
            Basal Metabolic Rate (BMR) or 1.22 times the female BMR.
            Default is set to 2.4 met for walking. Typical values include
            the following.
            _
            * 1 met = Metabolic rate of a resting seated person
            * 1.2 met = Metabolic rate of a standing person
            * 2.4 met = Metabolic rate of a person walking at ~1 m/s (2 mph)
        _clothing_: Data Collection or individual value of clothing insulation in clo.
            Default is set to 0.7 clo for long sleeve shirt and pants. Typical values
            include the following.
            _
            * 1 clo = Three-piece suit
            * 0.5 clo = Shorts + T-shirt
            * 0 clo = No clothing
        pet_par_: Optional comfort parameters from the "LB PET Body Parameters"
            component to specify the body properties of the human subject.
            This includes the age, height, sex, body mass, and more.
            The default attempts to model as average of a human body as
            possible.
        _run: Set to True to run the component.

    Returns:
        report: Reports, errors, warnings, etc.
        pet: Physiological Equivalent Temperature (PET) [C]. PET is a "feels like"
            temperature value and is defined as the operative temperature
            of a reference environment that would cause the same physiological
            response in the human subject as the environment under study.
            That is, the same skin temperature and core body temperature.
        core_temp: The core body temperature of the human subject [C].
        skin_temp: The skin temperature of the human subject underneath the
            clothing layer [C].
        comfort: Integers noting whether the input conditions result in no
            thermal stress (aka. they are comfortable.
            _
            Values are one of the following:
                * 0 = uncomfortable
                * 1 = comfortable
        condition: Integers noting the thermal status of a subject.
            _
            Values are one of the following:
                * -1 = cold
                *  0 = netural
                * +1 = hot
        category: Integers noting the thermal status on a nine-point scale.
            _
            Values are one of the following:
                * -4 = very strong/extreme cold stress
                * -3 = strong cold stress
                * -2 = moderate cold stress
                * -1 = slight cold stress
                * 0 = no thermal stress
                * +1 = slight heat stress
                * +2 = moderate heat stress
                * +3 = strong heat stress
                * +4 = very strong/extreme heat stress
        core_cond: Integers noting the classification of core body temperature.
            _
            Values are one of the following:
                * -2 = Hypothermia
                * -1 = Cold
                * 0 = Normal
                * +1 = Hot
                * +2 = Hyperthermia
        comf_obj: A Python object containing all inputs and results of the
            analysis.  This can be plugged into components like the "Comfort
            Statistics" component to get further information.
"""

ghenv.Component.Name = 'LB PET Comfort'
ghenv.Component.NickName = 'PET'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '1 :: Analyze Data'
ghenv.Component.AdditionalHelpFromDocStrings = '5'

try:
    from ladybug.datatype.temperature import Temperature
    from ladybug.datacollection import BaseCollection
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_comfort.collection.pet import PET
    from ladybug_comfort.parameter.pet import PETParameter
    from ladybug_comfort.pet import physiologic_equivalent_temperature, pet_category, \
        pet_category_humid, core_temperature_category
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_comfort:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


def extract_collections(input_list):
    """Process inputs into collections and floats."""
    defaults = [None, _air_temp, 0.1, None, 2.4, 0.7, 101325]
    data_colls = []
    for i, input in enumerate(input_list):
        if input is None:
            input_list[i] = defaults[i]
        elif isinstance(input, BaseCollection):
            data_colls.append(input)
        else:
            try:
                input_list[i] = float(input)
            except ValueError as e:
                raise TypeError('input {} is not valid. Expected float or '
                                'DataCollection. Got {}'.format(input, type(input)))
    return input_list, data_colls


if all_required_inputs(ghenv.Component) and _run is True:
    # Process inputs and assign defaults.
    input_list = [_air_temp, _mrt_, _air_speed_, _rel_humid,
                  _met_rate_, _clothing_, _pressure_]
    input, data_colls = extract_collections(input_list)
    pet_par = pet_par_ or PETParameter()

    if data_colls == []:
        # The inputs are all individual values.
        pet_result = physiologic_equivalent_temperature(
            input[0], float(input[1]), input[2], input[3], input[4], input[5],
            pet_par.age, pet_par.sex, pet_par.height, pet_par.body_mass,
            pet_par.posture, input[6])
        pet = pet_result['pet']
        core_temp = pet_result['t_core']
        skin_temp = pet_result['t_skin']
        cat_func = pet_category_humid if pet_par.humid_acclimated else pet_category
        category = cat_func(pet)
        core_cond = core_temperature_category(pet_result['t_core'])
        comfort = category == 0
        condition = 0
        if category < 0:
            condition = -1
        elif category > 0:
            condition = 1
    else:
        # The inputs include Data Collections.
        if not isinstance(_air_temp, BaseCollection):
            _air_temp = data_colls[0].get_aligned_collection(
                float(_air_temp), Temperature(), 'C')

        comf_obj = PET(
            _air_temp, _rel_humid, _mrt_, _air_speed_, _pressure_,
            _met_rate_, _clothing_, pet_par)
        pet = comf_obj.physiologic_equivalent_temperature
        core_temp = comf_obj.core_body_temperature
        skin_temp = comf_obj.skin_temperature
        comfort = comf_obj.is_comfortable
        condition = comf_obj.thermal_condition
        category = comf_obj.pet_category
        core_cond = comf_obj.core_temperature_category



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Generate screen-oriented text that displays in the Rhino scene as a head-up
display (HUD).
_
This is useful when there are certain summary results or information that should
always be displayed on-screen.
-

    Args:
        _text: Text string to be displayed in the plane of the screen.
        leg_par2d_: Optional 2D LegendParameters from the "LB Legend Parameters 2D"
            component, which will be used to customize a text in the plane
            of the screen. Note that only the text_height, origin_x and origin_y
            inputs of this component affect the placement of the text.
        _font_: An optional text string to specify the font to be used for the text.
            Examples include "Arial", "Times New Roman", "Courier" (all without
            quotations). Default is "Arial".
"""

from ghpythonlib.componentbase import executingcomponent as component
import Grasshopper, GhPython
import System
import Rhino
import rhinoscriptsyntax as rs

class MyComponent(component):
    
    def __init__(self):
        super(MyComponent,self).__init__()
        self.text_2d_args = None
    
    def RunScript(self, _text, leg_par2d_, _font_):
        ghenv.Component.Name = 'LB Screen Oriented Text'
        ghenv.Component.NickName = 'Text2D'
        ghenv.Component.Message = '1.6.0'
        ghenv.Component.Category = 'Ladybug'
        ghenv.Component.SubCategory = '4 :: Extra'
        ghenv.Component.AdditionalHelpFromDocStrings = '0'
        
        try:
            from ladybug.legend import Legend, Legend2DParameters
        except ImportError as e:
            raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))
        
        try:
            from ladybug_rhino.color import black
            from ladybug_rhino.grasshopper import all_required_inputs, longest_list
        except ImportError as e:
            raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))
        
        import scriptcontext as sc
        import Rhino.Geometry as rg
        
        if all_required_inputs(ghenv.Component):
            # get the screen dimensions
            active_view = sc.doc.Views.ActiveView.ActiveViewport
            v_size = active_view.Size
            vw = v_size.Width
            vh = v_size.Height
            _font = 'Arial' if _font_ is None else _font_
            
            # convert the inputs into arguments to be rendered
            self.text_2d_args = []
            for i, txt in enumerate(_text):
                if len(leg_par2d_) == 0:  # make default 2D parameters
                    l_par = Legend2DParameters()
                else:
                    l_par = longest_list(leg_par2d_, i)
                _height = Legend.parse_dim_2d(l_par.text_height, vh)
                or_x = Legend.parse_dim_2d(l_par.origin_x, vw)
                or_y = Legend.parse_dim_2d(l_par.origin_y, vh)
                d_args = (
                    txt, black(), rg.Point2d(or_x,or_y), False, _height, _font)
                self.text_2d_args.append(d_args)
        else:
            self.text_2d_args = None
    
    def DrawViewportWires(self, args):
        try:
            if self.text_2d_args is not None:
                # get the DisplayPipeline from the event arguments
                display = args.Display
                # render the 2D text from the arguments
                for draw_args in self.text_2d_args:
                    display.Draw2dText(*draw_args)
        except Exception, e:
            System.Windows.Forms.MessageBox.Show(str(e), "script error")



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>


"""
Create a set of parameters that define the body characteristics for the PET model.
-
These parameters can be plugged into any of the components that compute
PET thermal comfort.
-

    Args:
        _age_: The age of the human subject in years. (Default: 36 years for middle age
            of the average worldwide life expectancy).
        _sex_: A value between 0 and 1 to indicate the sex of the human subject,
            which influences the computation of basal metabolism. 0 indicates male.
            1 indicates female and any number in between denotes a weighted average
            between the two. (Default: 0.5).
        _height_: The height of the human subject in meters. Average male height
            is around 1.75m while average female height is 1.55m. (Default: 1.65m
            for a worldwide average between male and female height).
        _body_mass_: The body mass of the human subject in kilograms. (Default: 62 kg
            for the worldwide average adult human body mass).
        _posture_: A text string indicating the posture of the body. Letters must
            be lowercase. Default is "standing". Choose from the following:
            _
            * standing
            * seated
            * crouching
        humid_acclim_: A boolean to note whether the human subject is acclimated
            to a humid/tropical climate (True) or is acclimated to a temperate
            climate (False). When True, the categories developed by Lin and
            Matzarakis (2008) will be used to assess comfort instead of the original
            categories developed by Matzarakis and Mayer (1996).

    Returns:
        pet_par: A PET comfort parameter object that can be plugged into any of the
            components that compute PET thermal comfort.
"""

ghenv.Component.Name = 'LB PET Body Parameters'
ghenv.Component.NickName = 'PETPar'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '4 :: Extra'
ghenv.Component.AdditionalHelpFromDocStrings = '0'

try:
    from ladybug_comfort.parameter.pet import PETParameter
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_comfort:\n\t{}'.format(e))


pet_par = PETParameter(_age_, _sex_, _height_, _body_mass_, _posture_, humid_acclim_)



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Visualize the solar energy falling on different direction as a rose.
_
By default, the Radiation Rose depicts the amount of solar energy received
by a vertical wall facing each of the directions of the compass rose.
_
This is useful for understanding the radiation harm/benefit experienced by
different building orientations or the orientations with the highest peak cooling
load (for sky matrices of clear skies). The tilt_angle can be used to assess the
solar energy falling on geometries that are not perfectly vertical, such
as tilted photovoltaic panels.
-

    Args:
        _sky_mtx: A Sky Matrix from the "LB Cumulative Sky Matrix" component or the
            "LB Benefit Sky Matrix" component, which describes the radiation
            coming from the various patches of the sky.
        context_:  Rhino Breps and/or Rhino Meshes representing context geometry
            that can block solar radiation to the center of the radiation rose.
        _dir_count_: An integer greater than or equal to 3, which notes the number of
            arrows to be generated for the radiation rose. (Default: 36).
        _tilt_angle_: A number between 0 and 90 that sets the vertical tilt angle (aka.
            the altitude) for all of the directions. By default, the Radiation
            Rose depicts the amount of solar energy received by a vertical
            wall (tilt_angle=0). The tilt_angle can be changed to a specific
            value to assess the solar energy falling on geometries that are not
            perfectly vertical, such as a tilted photovoltaic panel. (Default: 0).
        _center_pt_: A point for the center of the radiation rose. (Default: (0, 0, 0))
        _scale_: A number to set the scale of the Radiation Rose. The default is 1,
            which corresponds to a radius of 100 meters in the current Rhino
            model's unit system.
        _arrow_scale_: A fractional number to note the scale of the radiation rose arrows
            in relation to the entire graphic. (Default: 1).
        max_rad_: An optional number to set the level of radiation or irradiance associated
            with the full radius of the rose. If unspecified, this is determined
            by the maximum level of radiation in the input data but a number
            can be specified here to fix this at a specific value. This is
            particularly useful when comparing different roses to one another.
        irradiance_: Boolean to note whether the radiation rose should be plotted with units
            of cumulative Radiation (kWh/m2) [False] or with units of average
            Irradiance (W/m2) [True]. (Default: False).
        show_comp_: Boolean to indicate whether only one rose with total radiation
            should be displayed (False) or three roses with the solar radiation
            components (total, direct, and diffuse) should be shown. (Default: False).
        legend_par_: An optional LegendParameter object to change the display of the
            Radiation Rose.

    Returns:
        report: ...
        mesh: A colored mesh of arrows, representing the intensity of radiation
            from different cardinal directions.
        compass: A set of circles, lines and text objects that mark the cardinal
            directions in relation to the radiation rose.
        orient_lines: A list of line segments marking the orientation of each arrow.
        legend: A legend showing the kWh/m2 or W/m2 values that correspond to the colors
            of the mesh.
        title: A text object for the title of the radiation rose.
        dir_vecs: A list of vectors for each of the directions the rose is facing.
            All vectors are unit vectors.
        dir_values: Radiation values for each of the rose directions in kWh/m2 or W/m2.
            This will be one list if show_comp_ is "False" and a list of 3
            lists (aka. a Data Tree) for total, direct, diffuse if show_comp_
            is "True".
"""

ghenv.Component.Name = 'LB Radiation Rose'
ghenv.Component.NickName = 'RadRose'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '2 :: Visualize Data'
ghenv.Component.AdditionalHelpFromDocStrings = '4'

try:
    from ladybug_geometry.geometry3d.pointvector import Point3D
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_geometry:\n\t{}'.format(e))

try:
    from ladybug.viewsphere import view_sphere
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_radiance.visualize.radrose import RadiationRose
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_radiance:\n\t{}'.format(e))

try:
    from ladybug_rhino.config import conversion_to_meters
    from ladybug_rhino.togeometry import to_point3d
    from ladybug_rhino.fromgeometry import from_point3d, from_vector3d, \
        from_linesegment3d, from_mesh3d
    from ladybug_rhino.intersect import join_geometry_to_mesh, intersect_mesh_rays
    from ladybug_rhino.fromobjects import legend_objects, compass_objects
    from ladybug_rhino.text import text_objects
    from ladybug_rhino.grasshopper import all_required_inputs, \
        objectify_output, list_to_data_tree
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


def translate_rose(lb_mesh, lb_orient_lines, lb_compass, graphic, title_txt):
    """Translate radiation rose geometry into a format suitable for Rhino.

    Args:
        lb_mesh: A ladybug Mesh3D for the rose.
        lb_compass: A ladybug Compass object.
        graphic: A GraphicContainer for the rose.
        title_txt: Text for title of the rose.

    Returns:
        rose_mesh: A Rhino colored mesh for the rose.
        rose_compass: Rhino objects for the rose compass.
        rose_legend:  Rhino objects for the rose legend.
        rose_title: A bake-able title for the rose.
    """
    # translate the rose visualization, including legend and compass
    rose_mesh = from_mesh3d(lb_mesh)
    rose_legend = legend_objects(graphic.legend)
    rose_angles = list(range(0, 360, int(360 / _dir_count_)))
    start, stop, step, rose_angles = 0, 360, 360 / _dir_count_, []
    while start < stop:
        rose_angles.append(start)
        start += step
    rose_angles = [int(n) for n in rose_angles]
    if len(rose_angles) > 36:
        rose_angles = rose_angles[::2]
    rose_compass = compass_objects(
        lb_compass, z, rose_angles, None, graphic.legend_parameters.font)
    orient_lines = [from_linesegment3d(l) for l in lb_orient_lines]

    # construct a title from the metadata
    rose_title = text_objects(title_txt, graphic.lower_title_location,
                              graphic.legend_parameters.text_height,
                              graphic.legend_parameters.font)

    return rose_mesh, orient_lines, rose_compass, rose_legend, rose_title


if all_required_inputs(ghenv.Component):
    # set defaults for global variables
    _dir_count_ = 36 if _dir_count_ is None else _dir_count_
    _tilt_angle_ = 0 if _tilt_angle_ is None else _tilt_angle_
    _arrow_scale_ = 1 if _arrow_scale_ is None else _arrow_scale_
    _scale_ = 1 if _scale_ is None else _scale_
    radius = (100 * _scale_) / conversion_to_meters()
    if _center_pt_ is not None:  # process the center point
        center_pt3d = to_point3d(_center_pt_)
        z = center_pt3d.z
    else:
        center_pt3d, z = Point3D(), 0

    # compute the intersection matrix if context is specified
    n_vecs = RadiationRose.radial_vectors(_dir_count_, _tilt_angle_)
    dir_vecs = [from_vector3d(vec) for vec in n_vecs]
    int_mtx = None
    if len(context_) != 0 and context_[0] is not None:
        shade_mesh = join_geometry_to_mesh(context_)
        p_vecs = view_sphere.tregenza_sphere_vectors if len(_sky_mtx.data[1]) == 145 \
            else view_sphere.reinhart_sphere_vectors
        patch_dirs = [from_vector3d(vec) for vec in p_vecs]
        int_mtx, angles = intersect_mesh_rays(
            shade_mesh, [from_point3d(center_pt3d)] * _dir_count_,
            patch_dirs, dir_vecs, 1)

    # create the RadiationRose object
    rad_rose = RadiationRose(
        _sky_mtx, int_mtx, _dir_count_, _tilt_angle_, legend_par_,
        irradiance_, center_pt3d, radius, _arrow_scale_)

    # create the rose visualization
    if not show_comp_:  # only create the total rose mesh
        mesh, orient_lines, compass, legend, title = \
            translate_rose(*rad_rose.draw(max_rad=max_rad_))
        dir_values = rad_rose.total_values
    else:  # create roses for total, direct and diffuse
        # loop through the 3 radiation types and produce a rose
        mesh, orient_lines, compass, legend, title = [], [], [], [], []
        rad_types = ('total', 'direct', 'diffuse')
        for rose_i in range(3):
            cent_pt = Point3D(center_pt3d.x + radius * 3 * rose_i,
                              center_pt3d.y, center_pt3d.z)
            rose_mesh, rose_lines, rose_compass, rose_legend, rose_title = \
                translate_rose(*rad_rose.draw(rad_types[rose_i], cent_pt, max_rad_))
            mesh.append(rose_mesh)
            compass.extend(rose_compass)
            orient_lines.extend(rose_lines)
            legend.extend(rose_legend)
            title.append(rose_title)
        dir_values = list_to_data_tree(
            (rad_rose.total_values, rad_rose.direct_values, rad_rose.diffuse_values))

    # output the visualization set
    vis_set = [rad_rose, max_rad_, show_comp_]
    vis_set = objectify_output('VisualizationSet Aruments [RadiationRose]', vis_set)



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Calculate the incident radiation on geometry using a sky matrix from the "Cumulative
Sky Matrix" component.
_
Such studies of incident radiation can be used to apprxomiate the energy that can
be collected from photovoltaic or solar thermal systems. They are also useful
for evaluating the impact of a building's orientation on both energy use and the
size/cost of cooling systems. For studies of photovoltaic potential or building
energy use impact, a sky matrix from EPW radiation should be used. For studies
of cooling system size/cost, a sky matrix derived from the STAT file's clear sky
radiation should be used.
_
NOTE THAT NO REFLECTIONS OF SOLAR ENERGY ARE INCLUDED
IN THE ANALYSIS PERFORMED BY THIS COMPONENT.
_
Ground reflected irradiance is crudely acounted for by means of an emissive
"ground hemisphere," which is like the sky dome hemisphere and is derived from
the ground reflectance that is associated with the connected _sky_mtx. This
means that including geometry that represents the ground surface will effectively
block such crude ground reflection.
_
Also note that this component uses the CAD environment's ray intersection methods,
which can be fast for geometries with low complexity but does not scale well for
complex geometries or many test points. For such complex cases and situations
where relfection of solar energy are important, honeybee-radiance should be used.
-

    Args:
        _sky_mtx: A Sky Matrix from the "LB Cumulative Sky Matrix" component or the
            "LB Benefit Sky Matrix" component, which describes the radiation
            coming from the various patches of the sky. The "LB Sky Dome"
            component can be used to visualize any sky matrix to understand
            its relationship to the test geometry.
        _geometry: Rhino Breps and/or Rhino Meshes for which incident radiation analysis
            will be conducted. If Breps are input, they will be subdivided using
            the _grid_size to yeild individual points at which analysis will
            occur. If a Mesh is input, radiation analysis analysis will be
            performed for each face of this mesh instead of subdividing it.
        context_: Rhino Breps and/or Rhino Meshes representing context geometry
            that can block solar radiation to the test _geometry.
        _grid_size: A positive number in Rhino model units for the size of grid
            cells at which the input _geometry will be subdivided for incident
            radiation analysis. The smaller the grid size, the higher the
            resolution of the analysis and the longer the calculation will take.
            So it is recommended that one start with a large value here and
            decrease the value as needed. However, the grid size should usually
            be smaller than the dimensions of the smallest piece of the _geometry
            and context_ in order to yield meaningful results.
        _offset_dist_: A number for the distance to move points from the surfaces
            of the input _geometry.  Typically, this should be a small positive
            number to ensure points are not blocked by the mesh. (Default: 10 cm
            in the equivalent Rhino Model units).
        irradiance_: Boolean to note whether the study should output units of cumulative
            Radiation (kWh/m2) [False] or units of average Irradiance (W/m2)
            [True].  (Default: False).
        legend_par_: Optional legend parameters from the "LB Legend Parameters"
            that will be used to customize the display of the results.
        _cpu_count_: An integer to set the number of CPUs used in the execution of the
            intersection calculation. If unspecified, it will automatically default
            to one less than the number of CPUs currently available on the
            machine or 1 if only one processor is available.
        _run: Set to "True" to run the component and perform incident radiation
            analysis.

    Returns:
        report: ...
        points: The grid of points on the test _geometry that are be used to perform
            the incident radiation analysis.
        results: A list of numbers that aligns with the points. Each number indicates
            the cumulative incident radiation received by each of the points
            from the sky matrix in kWh/m2.
        total: A number for the total incident solar energy falling on all input geometry
            in kWh. Note that, unlike the radiation results above, which are
            normlaized by area, these values are not area-normalized and so
            the input geometry must be represented correctly in the Rhino
            model's unit system in order for this output to be meaningful.
        mesh: A colored mesh of the test _geometry representing the cumulative
            incident radiation received by the input _geometry.
        legend: A legend showing the kWh/m2 that correspond to the colors of the mesh.
        title: A text object for the study title.
        int_mtx: A Matrix object that can be connected to the "LB Deconstruct Matrix"
            component to obtain detailed patch-by-patch results of the study.
            Each sub-list of the matrix (aka. branch of the Data Tree) represents
            one of the points used for analysis. The length of each sub-list
            matches the number of sky patches in the input sky matrix (145 for
            the default Tregenza sky and 577 for the high_density Reinhart sky).
            Each value in the sub-list is a value between 0 and 1 indicating the
            relationship between the point and the patch of the sky. A value of
            "0", indicates that the patch is not visible for that point at all
            while a value of "1" indicates that the patch hits the surface that
            the point represents head on.
"""

ghenv.Component.Name = "LB Incident Radiation"
ghenv.Component.NickName = 'IncidentRadiation'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '3 :: Analyze Geometry'
ghenv.Component.AdditionalHelpFromDocStrings = '1'

import math
try:  # python 2
    from itertools import izip as zip
except ImportError:  # python 3
    pass

try:
    from ladybug.viewsphere import view_sphere
    from ladybug.graphic import GraphicContainer
    from ladybug.legend import LegendParameters
    from ladybug.color import Colorset
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.config import conversion_to_meters
    from ladybug_rhino.togeometry import to_joined_gridded_mesh3d
    from ladybug_rhino.fromgeometry import from_mesh3d, from_point3d, from_vector3d
    from ladybug_rhino.fromobjects import legend_objects
    from ladybug_rhino.text import text_objects
    from ladybug_rhino.intersect import join_geometry_to_mesh, intersect_mesh_rays
    from ladybug_rhino.grasshopper import all_required_inputs, hide_output, \
        show_output, objectify_output, de_objectify_output, recommended_processor_count
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component) and _run:
    # set the default offset distance and _cpu_count
    _offset_dist_ = _offset_dist_ if _offset_dist_ is not None \
        else 0.1 / conversion_to_meters()
    workers = _cpu_count_ if _cpu_count_ is not None else recommended_processor_count()

    # create the gridded mesh from the geometry
    study_mesh = to_joined_gridded_mesh3d(_geometry, _grid_size)
    points = [from_point3d(pt.move(vec * _offset_dist_)) for pt, vec in
              zip(study_mesh.face_centroids, study_mesh.face_normals)]
    hide_output(ghenv.Component, 1)

    # mesh the geometry and context
    shade_mesh = join_geometry_to_mesh(_geometry + context_)

    # deconstruct the matrix and get the sky dome vectors
    mtx = de_objectify_output(_sky_mtx)
    total_sky_rad = [dir_rad + dif_rad for dir_rad, dif_rad in zip(mtx[1], mtx[2])]
    ground_rad = [(sum(total_sky_rad) / len(total_sky_rad)) * mtx[0][1]] * len(total_sky_rad)
    all_rad = total_sky_rad + ground_rad 
    lb_vecs = view_sphere.tregenza_dome_vectors if len(total_sky_rad) == 145 \
        else view_sphere.reinhart_dome_vectors
    if mtx[0][0] != 0:  # there is a north input for sky; rotate vectors
        north_angle = math.radians(mtx[0][0])
        lb_vecs = tuple(vec.rotate_xy(north_angle) for vec in lb_vecs)
    lb_grnd_vecs = tuple(vec.reverse() for vec in lb_vecs)
    all_vecs = [from_vector3d(vec) for vec in lb_vecs + lb_grnd_vecs]

    # intersect the rays with the mesh
    normals = [from_vector3d(vec) for vec in study_mesh.face_normals]
    int_matrix_init, angles = intersect_mesh_rays(
        shade_mesh, points, all_vecs, normals, cpu_count=workers)

    # compute the results
    results = []
    int_matrix = []
    for int_vals, angs in zip(int_matrix_init, angles):
        pt_rel = [ival * math.cos(ang) for ival, ang in zip(int_vals, angs)]
        int_matrix.append(pt_rel)
        rad_result = sum(r * w for r, w in zip(pt_rel, all_rad))
        results.append(rad_result)

    # convert to irradiance if requested
    study_name = 'Incident Radiation'
    if irradiance_:
        study_name = 'Incident Irradiance'
        factor = 1000 / _sky_mtx.wea_duration if hasattr(_sky_mtx, 'wea_duration') \
            else 1000 / (((mtx[0][3] - mtx[0][2]).total_seconds() / 3600) + 1)
        results = [r * factor for r in results]

    # output the intersection matrix and compute total radiation
    int_mtx = objectify_output('Geometry/Sky Intersection Matrix', int_matrix)
    unit_conv = conversion_to_meters() ** 2
    total = 0
    for rad, area in zip(results, study_mesh.face_areas):
        total += rad * area * unit_conv

    # create the mesh and legend outputs
    l_par = legend_par_ if legend_par_ is not None else LegendParameters()
    if hasattr(_sky_mtx, 'benefit_matrix') and _sky_mtx.benefit_matrix is not None:
        study_name = '{} Benefit/Harm'.format(study_name)
        if l_par.are_colors_default:
            l_par.colors = reversed(Colorset.benefit_harm())
        if l_par.min is None:
            l_par.min = min((min(results), -max(results)))
        if l_par.max is None:
            l_par.max = max((-min(results), max(results)))
    graphic = GraphicContainer(results, study_mesh.min, study_mesh.max, l_par)
    graphic.legend_parameters.title = 'kWh/m2' if not irradiance_ else 'W/m2'
    title = text_objects(
        study_name, graphic.lower_title_location,
        graphic.legend_parameters.text_height * 1.5,
        graphic.legend_parameters.font)

    # create all of the visual outputs
    study_mesh.colors = graphic.value_colors
    mesh = from_mesh3d(study_mesh)
    legend = legend_objects(graphic.legend)



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Use this component to change the colors, range, and display of any Ladybug legend
along with the corresponding colored mesh that the legend refers to.
-
The legend parameters from this component have more limitations than the normal
Legend Parameters. However, these legend parameters will do auto-categorization
of data, binning values into groups based on custom ranges.
-

    Args:
        _domain: A list of one or more numbers noting the bondaries of the data
            categories. For example, [100, 2000] creates three categories of
            (<100, 100-2000, >2000). Values must always be ordered from lowest
            to highest.
        _colors: An list of color objects with a length equal to the number of items
            in the domain + 1. These are used to color each of the categories of data.
        categories_: An optional list of text strings with a length equal to the
            colors. These will be used to name each of the categories in the legend.
            If None, the legend text will simply mark the numerical ranges of the
            categories. (Default: None).
        continuous_cols_: Boolean noting whether colors generated are continuous
            or discrete. If True, the colors generated from the corresponding
            legend will be in a continuous gradient. If False, they will be
            categorized in incremental groups according to the segment_count.
            (Default: False).
        continuous_leg_: Boolean. If True, the colors along the legend will be in
            a continuous gradient. If False, they will be categorized in
            incremental groups according to the number_of_segments.
            Default is False for depicting discrete categories.
        num_decimals_: An optional integer to set the number of decimal
            places for the numbers in the legend text. Default is 2.
        larger_smaller_: Boolean noting whether to include larger than and
            smaller than (> and <) values after the upper and lower legend segment
            text. Default is False.
        vert_or_horiz_: Boolean. If True, the legend mesh and text points
            will be generated vertically.  If False, they will genrate a
            horizontal legend. Default is True for a vertically-oriented legend.
        base_plane_: A Plane to note the starting point and orientation from
            where the legend will be genrated. The default is the world XY plane
            at origin (0, 0, 0).
        seg_height_: An optional number to set the height of each of the legend
            segments. Default is 1.
        seg_width_: An optional number to set the width of each of the legend
            segments. Default is 1 when legend is vertical. When horizontal, the
            default is (text_height * (number_decimal_places + 2)).
        text_height_: An optional number to set the size of the text in model units.
            Default is half of the segment_height.
        font_: An optional text string to specify the font to be used for the text.
            Examples include "Arial", "Times New Roman", "Courier" (all without
            quotations). Default is "Arial".
    
    Returns:
        leg_par: A legend parameter object that can be plugged into any of the
            Ladybug components with a legend.
"""

ghenv.Component.Name = 'LB Legend Parameters Categorized'
ghenv.Component.NickName = 'LegendParCategorized'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '4 :: Extra'
ghenv.Component.AdditionalHelpFromDocStrings = '0'


try:
    from ladybug.legend import LegendParametersCategorized
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.togeometry import to_plane
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    if base_plane_:
        base_plane_ = to_plane(base_plane_)
    if len(categories_) == 0:
        categories_ = None

    leg_par = LegendParametersCategorized(
        domain=_domain, colors=_colors, category_names=categories_,
        base_plane=base_plane_)

    leg_par.continuous_colors = continuous_cols_
    leg_par.continuous_legend = continuous_leg_
    leg_par.decimal_count = num_decimals_
    leg_par.include_larger_smaller = larger_smaller_
    leg_par.vertical = vert_or_horiz_
    leg_par.segment_height = seg_height_
    leg_par.segment_width = seg_width_
    leg_par.text_height = text_height_
    leg_par.font = font_



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Use this component to change the colors, numerical range, and/or number of divisions
of any Ladybug legend along with the corresponding colored mesh that the legend refers to.
-
Any Ladybug component that outputs a colored mesh and a legend will have an input
that can accept Legend Parameters from this component.
-

    Args:
        min_: A number to set the lower boundary of the legend. If None, the
            minimum of the values associated with the legend will be used.
        max_: A number to set the upper boundary of the legend. If None, the
            maximum of the values associated with the legend will be used.
        seg_count_: An interger representing the number of steps between
            the high and low boundary of the legend. The default is set to 11
            and any custom values input in here should always be greater than or
            equal to 2.
        colors_: An list of color objects. Default is Ladybug's original colorset.
        continuous_leg_: Boolean. If True, the colors along the legend will be in
            a continuous gradient. If False, they will be categorized in
            incremental groups according to the number_of_segments.
            Default is False for depicting discrete categories.
        num_decimals_: An optional integer to set the number of decimal
            places for the numbers in the legend text. Default is 2.
        larger_smaller_: Boolean noting whether to include larger than and
            smaller than (> and <) values after the upper and lower legend segment
            text. Default is False.
        vert_or_horiz_: Boolean. If True, the legend mesh and text points
            will be generated vertically.  If False, they will genrate a
            horizontal legend. Default is True for a vertically-oriented legend.
        base_plane_: A Plane to note the starting point and orientation from
            where the legend will be genrated. The default is the world XY plane
            at origin (0, 0, 0).
        seg_height_: An optional number to set the height of each of the legend
            segments. Default is 1.
        seg_width_: An optional number to set the width of each of the legend
            segments. Default is 1 when legend is vertical. When horizontal, the
            default is (text_height * (number_decimal_places + 2)).
        text_height_: An optional number to set the size of the text in model units.
            Default is half of the segment_height.
        font_: An optional text string to specify the font to be used for the text.
            Examples include "Arial", "Times New Roman", "Courier" (all without
            quotations). Default is "Arial".
    
    Returns:
        leg_par: A legend parameter object that can be plugged into any of the
            Ladybug components with a legend.
"""

ghenv.Component.Name = 'LB Legend Parameters'
ghenv.Component.NickName = 'LegendPar'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '4 :: Extra'
ghenv.Component.AdditionalHelpFromDocStrings = '1'


try:
    from ladybug.legend import LegendParameters
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.togeometry import to_plane
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if colors_ == []:
    colors_ = None
if base_plane_:
    base_plane_ = to_plane(base_plane_)

leg_par = LegendParameters(min=min_, max=max_, segment_count=seg_count_,
                           colors=colors_, base_plane=base_plane_)

leg_par.continuous_legend = continuous_leg_
leg_par.decimal_count = num_decimals_
leg_par.include_larger_smaller = larger_smaller_
leg_par.vertical = vert_or_horiz_
leg_par.segment_height = seg_height_
leg_par.segment_width = seg_width_
leg_par.text_height = text_height_
leg_par.font = font_



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Convert a DataCollection to SI values.
-

    Args:
        _data_ip: A DataCollection in IP (or other) units.

    Returns:
        data_si: The DataCollection in SI units.
"""

ghenv.Component.Name = "LB To SI"
ghenv.Component.NickName = 'ToSI'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '4 :: Extra'
ghenv.Component.AdditionalHelpFromDocStrings = '3'

try:
    from ladybug.datacollection import BaseCollection
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    assert isinstance(_data_ip, BaseCollection), \
        '_data_ip must be a Data Collection. Got {}.'.format(type(_data_ip))
    data_si = _data_ip.to_si()


# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Construct a Ladybug Matrix object from a Grasshopper Data Tree of values.
-

    Args:
        _values: A Grasshopper Data Tree of values to be merged into a matrix object.

    Returns:
        matrix: A Ladybug Matrix object encapsulating all of the input values.
"""

ghenv.Component.Name = "LB Construct Matrix"
ghenv.Component.NickName = '+Matrix'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '4 :: Extra'
ghenv.Component.AdditionalHelpFromDocStrings = '0'

try:
    from ladybug_rhino.grasshopper import all_required_inputs, objectify_output, \
        data_tree_to_list
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    python_mtx = [row[1] for row in data_tree_to_list(_values)]
    matrix = objectify_output('Matrix', python_mtx)


# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Visualize a wind profile curve for a given terrain type.
_
Wind profiles assist with understanding how wind speed decreases as one approaches
the ground or increases as one leaves the ground. 
_
By default, the wind profile output by this component will be an average over the
_met_wind_vel data collection (or it can be for a single meteorological wind
velocity for point-in-time studies).
_
If a met_wind_dir_ data collection is connected, the wind profile will point in
the direction of prevailing wind direction by default. A profile_dir_ can then
be connected to understand the average wind profile from a specific cardinal
direction (eg. NE).
-

    Args:
        north_: A number between -360 and 360 for the counterclockwise
            difference between the North and the positive Y-axis in degrees.
            90 is West and 270 is East. This can also be Vector for the
            direction to North. (Default: 0)
        _met_wind_vel: A data collection of meteorological wind speed measured at
            the _met_height_ with the _met_terrian [m/s]. Typically, this comes
            from the "LB Import EPW" component. This can also be a single number
            for the meteorological wind speed in m/s.
        met_wind_dir_: An optional number between 0 and 360 representing the degrees
            from north that the meteorological wind is blowing. 0 = North,
            90 = East, 180 = South, 270 = West. This can also a data collection
            of meteorological wind directions. in which case the wind profile
            will be oriented towards the prevailing wind (unless a profile_dir_
            is connected). When unspecified, the wind profile is simply drawn
            in the XY plane.
        profile_dir_: An optional text string representing the cardinal direction that
            the wind profile represents. This input only has an effect when a
            data collection is connected for met_wind_dir_. It will be used
            to draw a wind profile for only the hours of the data collection
            where the wind is blowing in the specified direction. This can also
            be an integer that codes for a particular orientation. Choose from
            the following.
            _
            0 = N
            1 = NE
            2 = E
            3 = SE
            4 = S
            5 = SW
            6 = W
            7 = NW
        _terrain_: Text string that sets the terrain class associated with the wind profile.
            This can also be an integer that codes for the terrain. (Default: city).
            Must be one the following.
            _
            0 = city - 50% of buildings above 21m over a distance of at least 2000m upwind.
            1 = suburban - suburbs, wooded areas.
            2 = country - open, with scattered objects generally less than 10m high.
            3 = water - flat areas downwind of a large water body (max 500m inland).
        _met_height_: A number for the height above the ground at which the meteorological
            wind speed is measured in meters. (Default: 10 meters, which is the
            standard used by most airports and EPW files).
        _met_terrain_: Text string that sets the terrain class associated with the
            meteorological wind speed. This can also be an integer that codes
            for the terrain. (Default: country, which is typical of most
            airports where wind measurements are taken). Must be one the following.
            _
            0 = city - 50% of buildings above 21m over a distance of at least 2000m upwind.
            1 = suburban - suburbs, wooded areas.
            2 = country - open, with scattered objects generally less than 10m high.
            3 = water - flat areas downwind of a large water body (max 500m inland).
        log_law_: A boolean to note whether the wind profile should use a logarithmic
            law to determine wind speeds instead of the default power law,
            which is used by EnergyPlus. (Default: False).
        _base_pt_: A point that sets the ground level frm which the wind profile is
            drawn. By default, the profile is generated at the scene origin.
        _profile_height_: A number in meters to specify the maximum height of the
            wind profile. (Default: 30 meters).
        _vec_spacing_: A number in meters to specify the difference in height between
            each of the mesh arrows. (Default 2 meters).
        _vec_scale_: A number to denote the length dimension of a 1 m/s wind vector
            in meters. This can be used to change the scale of the wind
            vector meshes in relation to the height of the wind profile
            curve. (Default: 5 meters).
        legend_par_: An optional LegendParameter object to change the display of the
            wind profile.

    Returns:
        report: Reports, errors, warnings, etc.
        wind_speeds: A list of wind speeds in [m/s] that correspond to the wind
            vectors slong the height of the wind profile visualization.
        wind_vectors: A list of vectors that built the profile. Note that the
            magnitude of these vectors is scaled based on the _vec_scale_
            input and a _vec_scale_ of 1 will make the magnitude of the
            vector equal to the wind speed in [m/s].
        anchor_pts: A list of anchor points for each of the vectors above, which
            correspond to the height above the ground for each of the vectors.
        mesh_arrows: A list of colored mesh objects that represent the wind speeds
            along the height of the wind profile.
        profile_curve: A curve outlining the wind speed as it changes with height.
        speed_axis: A list of line segments and text objects that mark the X axis,
            which relates to the wind speed in (m/s).
        height_axis: A list of line segments and text objects that mark the Y axis,
            which relates to the the height above the ground in Rhino model units.
        legend: A legend for the colored mesh_arrows, which notes their speed.
        title: A text object for the global_title.
        vis_set: An object containing VisualizationSet arguments for drawing a detailed
            version of the Wind Profile in the Rhino scene. This can be connected to
            the "LB Preview Visualization Set" component to display this version
            of the Wind Profile in Rhino.
"""

ghenv.Component.Name = 'LB Wind Profile'
ghenv.Component.NickName = 'WindProfile'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '2 :: Visualize Data'
ghenv.Component.AdditionalHelpFromDocStrings = '5'

import math

try:
    from ladybug_geometry.geometry2d import Vector2D
    from ladybug_geometry.geometry3d import Point3D
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_geometry:\n\t{}'.format(e))

try:
    from ladybug.datatype.speed import WindSpeed
    from ladybug.datacollection import BaseCollection
    from ladybug.graphic import GraphicContainer
    from ladybug.windprofile import WindProfile
    from ladybug.windrose import WindRose
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.togeometry import to_point3d, to_vector2d
    from ladybug_rhino.fromgeometry import from_point3d, from_vector3d, \
        from_mesh3d, from_linesegment3d, from_polyline3d
    from ladybug_rhino.text import text_objects
    from ladybug_rhino.fromobjects import legend_objects
    from ladybug_rhino.grasshopper import all_required_inputs, objectify_output
    from ladybug_rhino.config import conversion_to_meters, units_system
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))

# dictionary to map integers to terrain types
TERRAIN_TYPES = {
    '0': 'city',
    '1': 'suburban',
    '2': 'country',
    '3': 'water',
    'city': 'city',
    'suburban': 'suburban',
    'country': 'country',
    'water': 'water'
}

# dictionary to map integers to cardinal directions
DIR_TEXT = {
    '0': 'N', '1': 'NE', '2': 'E', '3': 'SE', '4': 'S', '5': 'SW', '6': 'W', '7': 'NW',
    'N': 'N', 'NE': 'NE', 'E': 'E', 'SE': 'SE', 'S': 'S', 'SW': 'SW', 'W': 'W', 'NW': 'NW'
}
DIR_RANGE = {
    'N': (337.5, 22.5), 'NE': (22.5, 67.5), 'E': (67.5, 112.5), 'SE': (112.5, 157.5),
    'S': (157.5, 202.5), 'SW': (202.5, 247.5), 'W': (247.5, 292.5), 'NW': (292.5, 337.5)
}


if all_required_inputs(ghenv.Component):
    # interpret the model units
    scale_fac = 1 / conversion_to_meters()
    unit_sys = units_system()

    # set default values
    if north_ is not None:  # process the north_
        try:
            north_ = math.degrees(
                to_vector2d(north_).angle_clockwise(Vector2D(0, 1)))
        except AttributeError:  # north angle instead of vector
            north_ = float(north_)
    else:
        north_ = 0
    _terrain_ = 'city' if _terrain_ is None else TERRAIN_TYPES[_terrain_.lower()]
    _met_height_ = 10 if _met_height_ is None else _met_height_
    _met_terrain_ = 'country' if _met_terrain_ is None \
        else TERRAIN_TYPES[_met_terrain_.lower()]
    log_law_ = False if log_law_ is None else log_law_
    bp = Point3D(0, 0, 0) if _base_pt_ is None else to_point3d(_base_pt_)
    if unit_sys in ('Feet', 'Inches'):
        _profile_height_ = 30.48 if _profile_height_ is None else _profile_height_
        _vec_spacing_ = 3.048 if _vec_spacing_ is None else _vec_spacing_
        feet_labels = True
    else:
        _profile_height_ = 30 if _profile_height_ is None else _profile_height_
        _vec_spacing_ = 2 if _vec_spacing_ is None else _vec_spacing_
        feet_labels = False
    _vec_scale_ = 5 if _vec_scale_ is None else _vec_scale_
    len_d, height_d = _vec_scale_, _vec_scale_ / 5

    # process the data collections and wind direction if reuqested
    if isinstance(met_wind_dir_, BaseCollection):
        if profile_dir_ is not None:
            dir_label = DIR_TEXT[profile_dir_]
            dir_txt = '\nWind Direction = {}'.format(dir_label)
        else:  # get the prevailing wind direction
            prev_dir = WindRose.prevailing_direction_from_data(met_wind_dir_, 8)[0]
            dir_label = DIR_TEXT[str(int(prev_dir / 45))]
            dir_txt = '\nPrevailing Wind Direction = {}'.format(dir_label)
        dir_range = DIR_RANGE[dir_label]
        met_wd = sum(dir_range) / 2 if dir_range != (337.5, 22.5) else 0
        if isinstance(_met_wind_vel, BaseCollection):
            lw, hg = dir_range
            if dir_range == (337.5, 22.5):
                pattern = [lw < v or v < hg for v in met_wind_dir_]
            else:
                pattern = [lw < v < hg for v in met_wind_dir_]
            _met_wind_vel = _met_wind_vel.filter_by_pattern(pattern)
    else:
        met_wd = float(met_wind_dir_) if met_wind_dir_ is not None else None
        dir_txt = '\nWind Direction = {} degrees'.format(int(met_wd)) \
            if met_wind_dir_ is not None else ''
    if isinstance(_met_wind_vel, BaseCollection):
        met_ws = _met_wind_vel.average
        head = _met_wind_vel.header
        loc_txt = '{} Terrain'.format(_terrain_.title()) if 'city' not in head.metadata \
            else '{} - {} Terrain'.format(head.metadata['city'], _terrain_.title())
        title_txt = '{}{}\nAverage Met Wind Speed = {} m/s'.format(
            loc_txt, dir_txt, round(met_ws, 2))
    else:
        met_ws = float(_met_wind_vel)
        title_txt = '{} Terrain{}\nMeteorological Speed = {} m/s'.format(
            _terrain_.title(), dir_txt, round(met_ws, 2))
    if met_wd is not None and north_ != 0:
        met_wd = met_wd - north_

    # create the wind profile and the graphic container
    profile = WindProfile(_terrain_, _met_terrain_, _met_height_, log_law_)
    _, mesh_ars, wind_speeds, wind_vectors, anchor_pts = \
        profile.mesh_arrow_profile(
            met_ws, _profile_height_, _vec_spacing_, met_wd, bp,
            len_d, height_d, scale_fac)
    profile_polyline, _, _ = profile.profile_polyline3d(
            met_ws, _profile_height_, 0.1,
            met_wd, bp, len_d, scale_fac)
    max_speed = round(wind_speeds[-1]) if _max_speed_ is None else _max_speed_
    max_pt = Point3D(bp.x + ((max_speed + 2) * len_d * scale_fac),
                     bp.y + (30 * scale_fac), bp.z)
    graphic = GraphicContainer(
        wind_speeds, bp, max_pt, legend_par_, WindSpeed(), 'm/s')

    # draw profile geometry and mesh arrows in the scene
    mesh_arrows = []
    for mesh, col in zip(mesh_ars, graphic.value_colors):
        mesh.colors = [col] * len(mesh)
        mesh_arrows.append(from_mesh3d(mesh))
    profile_curve = from_polyline3d(profile_polyline)

    # draw axes and legend in the scene
    txt_h = graphic.legend_parameters.text_height
    axis_line, axis_arrow, axis_ticks, text_planes, text = \
        profile.speed_axis(max_speed, met_wd, bp, len_d, scale_fac, txt_h)
    speed_axis = [from_linesegment3d(axis_line), from_mesh3d(axis_arrow)]
    for tic in axis_ticks:
        speed_axis.append(from_linesegment3d(tic))
    for i, (pl, txt) in enumerate(zip(text_planes, text)):
        txt_i_h = txt_h if i != len(text) - 1 else txt_h * 1.25
        txt_obj = text_objects(txt, pl, txt_i_h, graphic.legend_parameters.font, 1, 0)
        speed_axis.append(txt_obj)
    axis_line, axis_arrow, axis_ticks, text_planes, text = \
        profile.height_axis(_profile_height_, _vec_spacing_ * 2, met_wd, bp,
                            scale_fac, txt_h, feet_labels)
    height_axis = [from_linesegment3d(axis_line), from_mesh3d(axis_arrow)]
    for tic in axis_ticks:
        height_axis.append(from_linesegment3d(tic))
    for i, (pl, txt) in enumerate(zip(text_planes, text)):
        if i != len(text) - 1:
            txt_i_h, ha, va = txt_h, 2, 3
        else:
            txt_i_h, ha, va = txt_h * 1.25, 1, 5
        txt_obj = text_objects(txt, pl, txt_i_h, graphic.legend_parameters.font, ha, va)
        height_axis.append(txt_obj)
    
    # draw the legend and the title
    if graphic.legend_parameters.is_base_plane_default:
        graphic.legend_parameters.base_plane = \
            profile.legend_plane(max_speed, met_wd, bp, len_d, scale_fac)
    legend = legend_objects(graphic.legend)
    title_pl = profile.title_plane(met_wd, bp, len_d, scale_fac, txt_h)
    title = text_objects(title_txt, title_pl, txt_h, graphic.legend_parameters.font, 0, 0)

    # process the output lists of data
    anchor_pts = [from_point3d(pt) for pt in anchor_pts]
    wind_vectors = [from_vector3d(vec) for vec in wind_vectors]
    wind_speeds.insert(0, 0)  # insert 0 wind speed for bottom of curve

    # create the output VisualizationSet arguments
    vis_set = [profile, met_ws, met_wd, legend_par_, bp, _profile_height_,
               _vec_spacing_, len_d, height_d, max_speed, scale_fac, feet_labels]
    vis_set = objectify_output('VisualizationSet Aruments [WindProfile]', vis_set)



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Draw thermal comfort polygons on a Psychrometric Chart using the PMV model for
indoor thermal comfort.
_
This component can also plot passive strategy polygons on the psychrometric chart
and can compute the number of hours that lie inside each of the comfort / strategy
polygons.
-

    Args:
        _psyc_chart: A ladybug Psychrometric Chart object on which the PMV comfort
            polygons will be plot. This can be obtained from the "LB Psychrometric
            Chart" component.
        _mrt_: A number or list of numbers for the mean radiant temperature. These
            should be in Celsius if the Psychrometric Chart is in SI and
            Farenheit if the Psychrometric Chart is in IP. If None, a polygon for
            operative temperature will be plot, assuming that radiant temperature
            and air temperature are the same. (Default: None).
        _air_speed_: A number or list of numbers for the air speed values in m/s. If None, a
            low air speed of 0.1 m/s wil be used for all polygons. (Default: None).
        _met_rate_: A number or list of numbers for the metabolic rate in met. If None, a met
            rate of 1.1 met will be used for all polygons, indicating a human
            subject who is seated, typing. (Default: None).
        _clothing_: A number or list of numbers for the clothing level in clo. If None, a clo
            level of 0.7 clo will be used for all polygons, indicating a human
            subject with a long sleeve shirt and pants. (Default: None).
        pmv_par_: Optional PMVParameter object to specify parameters under
            which conditions are considered acceptable. If None, default will
            assume a PPD threshold of 10%, no absolute humidity constraints
            and a still air threshold of 0.1 m/s.
        merge_poly_: Boolean to note whether all comfort polygons should be merged
            into a single polygon instead of separate polygons for each set of
            input conditions. (Default: False).
        strategies_: An optional text input of passive strategies to be plot on the
            psychrometric chart as polygons.  It is recommended that the
            "LB Passive Strategy" component be used to select which polygons
            to plot.
        strategy_par_: Optional passive strategy parameters from the "LB Passive Strategy
            Parameters" component. This can be used to adjust various inputs
            used to generate strategy polygons including the maximum comfortable
            air speed, the building balance temperature, and the temperature
            limits for thermal mass and night flushing.
        solar_data_: An annual hourly continuous data collection of irradiance
            (or radiation) in W/m2 (or Wh/m2) that aligns with the data
            points on the psychrometric chart. This is only required when
            plotting a "Passive Solar Heating" strategy polygon on the chart.
            The irradiance values should be incident on the orientation of
            the passive solar heated windows. So using global horizontal
            radiation assumes that all windows are skylights (like a
            greenhouse). The "LB Directional Irradiance" component can be
            used to get irradiance data for a specific surface orientation.

    Returns:
        report: ...
        total_comfort: The percent of the data on the psychrometric chart that
            are inside all comfort and passive strategy polygons.
        total_comf_data: Data collection or a 0/1 value noting whether each of the data
            points on the psychrometric chart lies inside of a comfort polygon
            or a strategy polygon.
            _
            This can be connected to the "LB Create Legend" component to generate
            a list of colors that can be used to color the points output from
            "LB Psychrometric Chart" component to see exactly which points are
            comfortable and which are not.
            _
            Values are one of the following:
                0 = uncomfortable
                1 = comfortable
        polygon_names:  A list of names for each of the polygons. This will include both the
            comfort polygons and the passive strategy polygons. The order of these
            names correspondsto the total_strategies and strategies_data outputs.
        polygon_comfort: The percent of the input data that are in each of the comfort
            or passive strategy polygons. Each number here corresponds to the names
            in the polygon_names output above.
        polygon_data: A list of data collections or 0/1 values indicating whether each 
            of the data points on the psychrometric chart lies inside each of
            the comfort or a strategy polygons. Each data collection or here corresponds
            to the names in the polygon_names output above.
            _
            Values are one of the following:
                0 = uncomfortable
                1 = comfortable
        comfort_poly: Brep representing the range of comfort for the input mrt, air speed,
            metabolic rate and clothing level. IF multiple values have been
            input, multiple polygons will be output here.
        strategy_poly: Brep representing the area of the chart made comfortable by any
            input passive strategies. If multiple strategies have been input to
            the strategies_ input, multiple polygons will be output here.
        strategy_par_: Optional passive strategy parameters from the "LB Passive Strategy
            Parameters" component. This can be used to adjust assumptions
            underlying the passive strategy polygons such as the maximum
            comfortable fan speed, the building balance temperature, the
            temperature limits for night flushing, etc.
"""

ghenv.Component.Name = 'LB PMV Polygon'
ghenv.Component.NickName = 'PMV Polygon'
ghenv.Component.Message = '1.6.2'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '2 :: Visualize Data'
ghenv.Component.AdditionalHelpFromDocStrings = '2'

try:
    from ladybug.psychchart import PsychrometricChart
    from ladybug.datacollection import BaseCollection
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_comfort.chart.polygonpmv import PolygonPMV
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_comfort:\n\t{}'.format(e))

try:
    from ladybug_rhino.config import tolerance
    from ladybug_rhino.fromgeometry import from_polyline2d_to_offset_brep
    from ladybug_rhino.grasshopper import all_required_inputs, \
        list_to_data_tree, give_warning, de_objectify_output
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


def strategy_warning(polygon_name):
    """Give a warning about a polygon not fitting on the chart."""
    msg = 'Polygon "{}" could not fit on the chart given the current location of ' \
        'the comfort polygon(s).\nTry moving the comfort polygon(s) by changing ' \
        'its criteria to see the missing polygon.'.format(polygon_name)
    give_warning(ghenv.Component, msg)
    print(msg)


def process_polygon(polygon_name, polygon):
    """Process a strategy polygon that does not require any special treatment."""
    if polygon is not None:
        polygon_names.append(polygon_name)
        strategy_poly.append(from_polyline2d_to_offset_brep(polygon, offset, z))
        dat = poly_obj.evaluate_polygon(polygon, tolerance)
        dat = dat[0] if len(dat) == 1 else poly_obj.create_collection(dat, polygon_name)
        polygon_data.append(dat)
    else:
        strategy_warning(polygon_name)


def merge_polygon_data(poly_data):
    """Merge an array of polygon comfort conditions into a single data list."""
    val_mtx = [dat.values for dat in poly_data]
    merged_values = []
    for hr_data in zip(*val_mtx):
        hr_val = 1 if 1 in hr_data else 0
        merged_values.append(hr_val)
    return merged_values


if all_required_inputs(ghenv.Component):
    # unpack any passive strategy parameters
    if strategy_par_ is None:
        day_above, night_below, fan_spd, bal_temp, solar_cap, tim_c = \
            (12.0, 3.0, 1.0, 12.8, 50.0, 8)
    else:
        day_above, night_below, fan_spd, bal_temp, solar_cap, tim_c = \
            de_objectify_output(strategy_par_)

    # convert the temperature values to C if the polygon is in IP
    assert isinstance(_psych_chart, PsychrometricChart), 'PolygonPMV ' \
        'psychrometric chart must be a ladybug PsychrometricChart. ' \
        'Got {}.'.format(type(_psych_chart))
    z = _psych_chart.z
    offset = _psych_chart.x_dim * 0.25
    if _psych_chart.use_ip:
        _mrt_ = PolygonPMV.TEMP_TYPE.to_unit(_mrt_, 'C', 'F')

    # create the PolygonPMV object
    poly_obj = PolygonPMV(_psych_chart, _mrt_, _air_speed_, _met_rate_, _clothing_,
                          comfort_parameter=pmv_par_)

    # draw the comfort polygon
    polygon_names = []
    polygon_data = []
    if merge_poly_:
        comfort_poly = from_polyline2d_to_offset_brep(
            poly_obj.merged_comfort_polygon, offset, z)
        polygon_data.append(poly_obj.merged_comfort_data)
        polygon_names.append('Comfort')
    else:
        comfort_poly = [from_polyline2d_to_offset_brep(poly, offset, z)
                        for poly in poly_obj.comfort_polygons]
        polygon_data.extend(poly_obj.comfort_data)
        if len(comfort_poly) == 1:
            polygon_names.append('Comfort')
        else:
            names = ('Comfort {}'.format(i + 1) for i in range(len(comfort_poly)))
            polygon_names.extend(names)

    # draw passive strategies if requested
    if len(strategies_) != 0:
        strategy_poly = []
        if 'Evaporative Cooling' in strategies_:
            ec_poly = poly_obj.evaporative_cooling_polygon()
            process_polygon('Evaporative Cooling', ec_poly)
        if 'Mass + Night Vent' in strategies_:
            nf_poly = poly_obj.night_flush_polygon(day_above)
            p_name = 'Mass + Night Vent'
            if nf_poly is not None:
                polygon_names.append(p_name)
                strategy_poly.append(from_polyline2d_to_offset_brep(nf_poly, offset, z))
                dat = poly_obj.evaluate_night_flush_polygon(
                    nf_poly, _psych_chart.original_temperature, night_below, tim_c, tolerance)
                dat = dat[0] if len(dat) == 1 else poly_obj.create_collection(dat, p_name)
                polygon_data.append(dat)
            else:
                strategy_warning(p_name)
        if 'Occupant Use of Fans' in strategies_:
            fan_poly = poly_obj.fan_use_polygon(fan_spd)
            process_polygon('Occupant Use of Fans', fan_poly)
        if 'Capture Internal Heat' in strategies_:
            iht_poly = poly_obj.internal_heat_polygon(bal_temp)
            process_polygon('Capture Internal Heat', iht_poly)
        if 'Passive Solar Heating' in strategies_:
            p_name = 'Passive Solar Heating'
            if solar_data_ is None:
                msg = 'In order to plot a "{}" polygon, solar_data_ ' \
                    'must be plugged into this component.'.format(p_name)
                print(msg)
                give_warning(ghenv.Component, msg)
            else:
                bal_t = bal_temp if 'Capture Internal Heat' in strategies_ else None
                dat, delta = poly_obj.evaluate_passive_solar(
                    solar_data_, solar_cap, tim_c, bal_t)
                sol_poly = poly_obj.passive_solar_polygon(delta, bal_t)
                if sol_poly is not None:
                    polygon_names.append(p_name)
                    strategy_poly.append(from_polyline2d_to_offset_brep(sol_poly, offset, z))
                    dat = dat[0] if len(dat) == 1 else poly_obj.create_collection(dat, p_name)
                    polygon_data.append(dat)
                else:
                    strategy_warning(p_name)

    # compute total comfort values
    polygon_comfort = [dat.average * 100 for dat in polygon_data] if \
        isinstance(polygon_data[0], BaseCollection) else \
        [dat * 100 for dat in polygon_data]
    if isinstance(polygon_data[0], BaseCollection):
        merged_vals = merge_polygon_data(polygon_data)
        total_comf_data = poly_obj.create_collection(merged_vals, 'Total Comfort')
        total_comfort = total_comf_data.average * 100
    else:
        total_comf_data = 1 if sum(polygon_data) > 0 else 0
        total_comfort = total_comf_data * 100



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Deconstruct a Ladybug Matrix object into a Grasshopper Data Tree of values.
-

    Args:
        _matrix: A Ladybug Matrix object such as the intersection matrices output
            from any of the ray-tracing components (eg. "LB Direct Sun Hours").

    Returns:
        values: The numerical values of the matrix as a Grasshopper Data Tree.
"""

ghenv.Component.Name = "LB Deconstruct Matrix"
ghenv.Component.NickName = 'XMatrix'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '4 :: Extra'
ghenv.Component.AdditionalHelpFromDocStrings = '0'

try:
    from ladybug_rhino.grasshopper import all_required_inputs, de_objectify_output, \
        list_to_data_tree, merge_data_tree
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    values = []
    for i, mtx in enumerate(_matrix):
        values.append(list_to_data_tree(de_objectify_output(mtx), root_count=i))
    values = merge_data_tree(values)



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2023, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Construct a Ladybug DataType to be used in the header of a ladybug DataCollection.
-

    Args:
        _name: A name for the data type as a string.
        _unit: A unit for the data type as a string.
        cumulative_: Boolean to tell whether the data type can be cumulative when it
            is represented over time (True) or it can only be averaged over time
            to be meaningful (False).
        categories_: An optional list of text for categories to be associated with
            the data type. These categories will show up in the legend whenever
            data with this data type is visualized. The input should be
            text strings with a category number (integer) and name separated
            by a colon. For example:
            _
            .    -1: Cold
            .     0: Neutral
            .     1: Hot
    
    Returns:
        type: A Ladybug DataType object that can be assigned to the header
            of a Ladybug DataCollection.
"""

ghenv.Component.Name = "LB Construct Data Type"
ghenv.Component.NickName = 'ConstrType'
ghenv.Component.Message = '1.6.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '1 :: Analyze Data'
ghenv.Component.AdditionalHelpFromDocStrings = '1'

try:
    from ladybug.datatype.generic import GenericType
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # process the categories_ if they are supplied
    unit_descr = None
    if categories_ != []:
        unit_descr = {}
        for prop in categories_:
            key, value = prop.split(':')
            unit_descr[int(key)] = value.strip()

    if cumulative_:
        type = GenericType(_name, _unit, unit_descr=unit_descr,
                           point_in_time=False, cumulative=True)
    else:
        type = GenericType(_name, _unit, unit_descr=unit_descr)
